# Copyright: (c) 2019, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from collections import OrderedDict

from smbprotocol.reparse_point import ReparseTags
from smbprotocol.structure import (
    BoolField,
    BytesField,
    DateTimeField,
    EnumField,
    FlagField,
    IntField,
    Structure,
    StructureField,
    TextField,
    UuidField,
)


class AlignmentRequirement:
    """
    [MS-FSCC] 2.4.3 FileAlignmentInformation - AlignmentRequirement
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/9b0b9971-85aa-4651-8438-f1c4298bcb0d
    """

    FILE_BYTE_ALIGNMENT = 0x00000000
    FILE_WORD_ALIGNMENT = 0x00000001
    FILE_LONG_ALIGNMENT = 0x00000003
    FILE_QUAD_ALIGNMENT = 0x00000007
    FILE_OCTA_ALIGNMENT = 0x0000000F
    FILE_32_BYTE_ALIGNMENT = 0x0000001F
    FILE_64_BYTE_ALIGNMENT = 0x0000003F
    FILE_128_BYTE_ALIGNMENT = 0x0000007F
    FILE_256_BYTE_ALIGNMENT = 0x000000FF
    FILE_512_BYTE_ALIGNMENT = 0x000001FF


class InfoType:
    """
    [MS-SMB2] 2.2.39 SMB2 SET_INFO Request - InfoType
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/ee9614c4-be54-4a3c-98f1-769a7032a0e4

    The type of information being set.
    """

    SMB2_0_INFO_FILE = 0x01
    SMB2_0_INFO_FILESYSTEM = 0x02
    SMB2_0_INFO_SECURITY = 0x03
    SMB2_0_INFO_QUOTA = 0x04


class ModeInformation:
    """
    [MS-FSCC] 2.4.24 FileModeInformation - Mode
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/52df7798-8330-474b-ac31-9afe8075640c
    """

    FILE_WRITE_THROUGH = 0x00000002
    FILE_SEQUENTIAL_ONLY = 0x00000004
    FILE_NO_INTERMEDIATE_BUFFERING = 0x00000008
    FILE_SYNCHRONOUS_IO_ALERT = 0x00000010
    FILE_SYNCHRONOUS_IO_NONALERT = 0x00000020
    FILE_DELETE_ON_CLOSE = 0x00001000


class FileAttributes:
    """
    [MS-FSCC]

    2.6 File Attributes
    Combination of file attributes for a file or directory
    """

    FILE_ATTRIBUTE_ARCHIVE = 0x00000020
    FILE_ATTRIBUTE_COMPRESSED = 0x00000800
    FILE_ATTRIBUTE_DIRECTORY = 0x00000010
    FILE_ATTRIBUTE_ENCRYPTED = 0x00004000
    FILE_ATTRIBUTE_HIDDEN = 0x00000002
    FILE_ATTRIBUTE_NORMAL = 0x00000080
    FILE_ATTRIBUTE_NOT_CONTENT_INDEXED = 0x00002000
    FILE_ATTRIBUTE_OFFLINE = 0x00001000
    FILE_ATTRIBUTE_READONLY = 0x00000001
    FILE_ATTRIBUTE_REPARSE_POINT = 0x00000400
    FILE_ATTRIBUTE_SPARSE_FILE = 0x00000200
    FILE_ATTRIBUTE_SYSTEM = 0x00000004
    FILE_ATTRIBUTE_TEMPORARY = 0x00000100
    FILE_ATTRIBUTE_INTEGRITY_STREAM = 0x00008000
    FILE_ATTRIBUTE_NO_SCRUB_DATA = 0x00020000
    FILE_ATTRIBUTE_RECALL_ON_OPEN = 0x00040000
    FILE_ATTRIBUTE_PINNED = 0x00080000
    FILE_ATTRIBUTE_UNPINNED = 0x00100000
    FILE_ATTRIBUTE_RECALL_ON_DATA_ACCESS = 0x00400000


class FileInformationClass:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.33 SMB2 QUERY_DIRECTORY Request FileInformationClass
    2.2.37 SMB2 QUERY_INFO Request FileInformationClass
    2.2.39 SMB2 SET_INFO Request FileInformationClass
    Describe the format the data MUST be returned in. The format structure must
    is specified in https://msdn.microsoft.com/en-us/library/cc232064.aspx
    """

    FILE_NONE = 0
    FILE_DIRECTORY_INFORMATION = 1
    FILE_FULL_DIRECTORY_INFORMATION = 2
    FILE_BOTH_DIRECTORY_INFORMATION = 3
    FILE_BASIC_INFORMATION = 4
    FILE_STANDARD_INFORMATION = 5
    FILE_INTERNAL_INFORMATION = 6
    FILE_EA_INFORMATION = 7
    FILE_ACCESS_INFORMATION = 8
    FILE_RENAME_INFORMATION = 10
    FILE_LINK_INFORMATION = 11
    FILE_NAMES_INFORMATION = 12
    FILE_DISPOSITION_INFORMATION = 13
    FILE_POSITION_INFORMATION = 14
    FILE_FULL_EA_INFORMATION = 15
    FILE_MODE_INFORMATION = 16
    FILE_ALIGNMENT_INFORMATION = 17
    FILE_ALL_INFORMATION = 18
    FILE_ALLOCATION_INFORMATION = 19
    FILE_END_OF_FILE_INFORMATION = 20
    FILE_ALTERNATE_NAME_INFORMATION = 21
    FILE_STREAM_INFORMATION = 22
    FILE_PIPE_INFORMATION = 23
    FILE_PIPE_LOCAL_INFORMATION = 24
    FILE_PIPE_REMOTE_INFORMATION = 25
    FILE_COMPRESSION_INFORMATION = 28
    FILE_QUOTE_INFORMATION = 32
    FILE_NETWORK_OPEN_INFORMATION = 34
    FILE_ATTRIBUTE_TAG_INFORMATION = 35
    FILE_ID_BOTH_DIRECTORY_INFORMATION = 37
    FILE_ID_FULL_DIRECTORY_INFORMATION = 38
    FILE_VALID_DATA_LENGTH_INFORMATION = 39
    FILE_SHORT_NAME_INFORMATION = 40
    FILE_NORMALIZED_NAME_INFORMATION = 48


class FileSystemInformationClass:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.33 SMB2 QUERY_DIRECTORY Request FileInformationClass
    2.2.37 SMB2 QUERY_INFO Request FileInformationClass
    2.2.39 SMB2 SET_INFO Request FileInformationClass

    Describe the format the data for a file system information class MUST be
    returned in. The format structure is specified in
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/ee12042a-9352-46e3-9f67-c094b75fe6c3
    """

    FILE_FS_VOLUME_INFORMATION = 1
    FILE_FS_LABEL_INFORMATION = 2
    FILE_FS_SIZE_INFORMATION = 3
    FILE_FS_DEVICE_INFORMATION = 4
    FILE_FS_ATTRIBUTE_INFORMATION = 5
    FILE_FS_CONTROL_INFORMATION = 6
    FILE_FS_FULL_SIZE_INFORMATION = 7
    FILE_FS_OBJECT_ID_INFORMATION = 8
    FILE_FS_DRIVER_PATH_INFORMATION = 9
    FILE_FS_VOLUME_FLAGS_INFORMATION = 10
    FILE_FS_SECTOR_SIZE_INFORMATION = 11


class QueryInfoFlags:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.37 SMB2 QUERY_INFO Request Flags
    """

    NONE = 0x00000000
    SL_RESTART_SCAN = 0x00000001
    SL_RETURN_SINGLE_ENTRY = 0x00000002
    SL_INDEX_SPECIFIED = 0x00000004


class FileNameInformation(Structure):
    """
    [MS-FSCC] 2.1.7 FILE_NAME_INFORMATION
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/20406fb1-605f-4629-ba9a-c67ee25f23d2
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "file_name_length",
                    IntField(
                        size=4,
                        default=lambda s: len(s["file_name"]),
                    ),
                ),
                (
                    "file_name",
                    TextField(
                        encoding="utf-16-le",
                        size=lambda s: s["file_name_length"].get_value(),
                        default="",
                    ),
                ),
            ]
        )
        super().__init__()


class FileAccessInformation(Structure):
    """
    [MS-FSSC] 2.4.1 FileAccessInformation
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/01cf43d2-deb3-40d3-a39b-9e68693d7c90
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_ACCESS_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                # Technically a flag but it depends on the file type in question so keep as a raw int.
                ("access_flags", IntField(size=4)),
            ]
        )
        super().__init__()


class FileAllInformation(Structure):
    """
    [MS-FSSC] 2.4.2 FileAllInformation
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/95f3056a-ebc1-4f5d-b938-3f68a44677a6
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_ALL_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "basic_information",
                    StructureField(
                        structure_type=FileBasicInformation,
                        default=FileBasicInformation(),
                    ),
                ),
                (
                    "standard_information",
                    StructureField(
                        structure_type=FileStandardInformation,
                        default=FileStandardInformation(),
                    ),
                ),
                (
                    "internal_information",
                    StructureField(
                        structure_type=FileInternalInformation,
                        default=FileInternalInformation(),
                    ),
                ),
                (
                    "ea_information",
                    StructureField(
                        structure_type=FileEaInformation,
                        default=FileEaInformation(),
                    ),
                ),
                (
                    "access_information",
                    StructureField(
                        structure_type=FileAccessInformation,
                        default=FileAccessInformation(),
                    ),
                ),
                (
                    "position_information",
                    StructureField(
                        structure_type=FilePositionInformation,
                        default=FilePositionInformation(),
                    ),
                ),
                (
                    "mode_information",
                    StructureField(
                        structure_type=FileModeInformation,
                        default=FileModeInformation(),
                    ),
                ),
                (
                    "alignment_information",
                    StructureField(
                        structure_type=FileAlignmentInformation,
                        default=FileAlignmentInformation(),
                    ),
                ),
                (
                    "name_information",
                    StructureField(
                        structure_type=FileNameInformation,
                        default=FileNameInformation(),
                    ),
                ),
            ]
        )
        super().__init__()


class FileAlignmentInformation(Structure):
    """
    [MS-FSCC] 2.4.3 FileAlignmentInformation
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/9b0b9971-85aa-4651-8438-f1c4298bcb0d
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_ALIGNMENT_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "alignment_requirement",
                    EnumField(
                        size=4,
                        enum_type=AlignmentRequirement,
                    ),
                ),
            ]
        )
        super().__init__()


class FileAttributeTagInformation(Structure):
    """
    [MS-FSCC] 2.4.6 FileAttributeTagInformation
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/d295752f-ce89-4b98-8553-266d37c84f0e
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_ATTRIBUTE_TAG_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "file_attributes",
                    FlagField(
                        size=4,
                        flag_type=FileAttributes,
                        flag_strict=False,
                    ),
                ),
                (
                    "reparse_tag",
                    EnumField(
                        size=4,
                        enum_type=ReparseTags,
                        enum_strict=False,
                    ),
                ),
            ]
        )
        super().__init__()


class FileBasicInformation(Structure):
    """
    [MS-FSCC] 2.4.7 FileBasicInformation
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/16023025-8a78-492f-8b96-c873b042ac50
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_BASIC_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("creation_time", IntField(size=8)),
                ("last_access_time", IntField(size=8)),
                ("last_write_time", IntField(size=8)),
                ("change_time", IntField(size=8)),
                (
                    "file_attributes",
                    FlagField(
                        size=4,
                        flag_type=FileAttributes,
                        flag_strict=False,
                    ),
                ),
                ("reserved", IntField(size=4)),
            ]
        )
        super().__init__()


class FileBothDirectoryInformation(Structure):
    """
    [MS-FSCC] 2.4.8 FileBothDirectoryInformation
    https://msdn.microsoft.com/en-us/library/cc232095.aspx
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_BOTH_DIRECTORY_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("next_entry_offset", IntField(size=4)),
                ("file_index", IntField(size=4)),
                ("creation_time", DateTimeField(size=8)),
                ("last_access_time", DateTimeField(size=8)),
                ("last_write_time", DateTimeField(size=8)),
                ("change_time", DateTimeField(size=8)),
                ("end_of_file", IntField(size=8)),
                ("allocation_size", IntField(size=8)),
                (
                    "file_attributes",
                    FlagField(
                        size=4,
                        flag_type=FileAttributes,
                        flag_strict=False,
                    ),
                ),
                ("file_name_length", IntField(size=4, default=lambda s: len(s["file_name"]))),
                ("ea_size", IntField(size=4)),
                ("short_name_length", IntField(size=1, default=lambda s: len(s["short_name"]))),
                ("reserved", IntField(size=1)),
                ("short_name", BytesField(size=lambda s: s["short_name_length"].get_value())),
                (
                    "short_name_padding",
                    BytesField(
                        size=lambda s: 24 - len(s["short_name"]),
                        default=lambda s: b"\x00" * (24 - len(s["short_name"])),
                    ),
                ),
                ("file_name", BytesField(size=lambda s: s["file_name_length"].get_value())),
            ]
        )
        super().__init__()


class FileDirectoryInformation(Structure):
    """
    [MS-FSCC] 2.4.10 FileDirectoryInformation
    https://msdn.microsoft.com/en-us/library/cc232097.aspx
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_DIRECTORY_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("next_entry_offset", IntField(size=4)),
                ("file_index", IntField(size=4)),
                ("creation_time", DateTimeField(size=8)),
                ("last_access_time", DateTimeField(size=8)),
                ("last_write_time", DateTimeField(size=8)),
                ("change_time", DateTimeField(size=8)),
                ("end_of_file", IntField(size=8)),
                ("allocation_size", IntField(size=8)),
                (
                    "file_attributes",
                    FlagField(
                        size=4,
                        flag_type=FileAttributes,
                        flag_strict=False,
                    ),
                ),
                ("file_name_length", IntField(size=4, default=lambda s: len(s["file_name"]))),
                ("file_name", BytesField(size=lambda s: s["file_name_length"].get_value())),
            ]
        )
        super().__init__()


class FileDispositionInformation(Structure):
    """
    [MS-FSCC] 2.4.11 FileDispositionInformation
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/12c3dd1c-14f6-4229-9d29-75fb2cb392f6
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_DISPOSITION_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("delete_pending", BoolField(size=1)),
            ]
        )
        super().__init__()


class FileEaInformation(Structure):
    """
    [MS-FSCC] 2.4.12 FileEaInformation
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/db6cf109-ead8-441a-b29e-cb2032778b0f
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_EA_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("ea_size", IntField(size=4)),
            ]
        )
        super().__init__()


class FileEndOfFileInformation(Structure):
    """
    [MS-FSCC] 2.4.13 FileEndOfFileInformation
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/75241cca-3167-472f-8058-a52d77c6bb17
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_END_OF_FILE_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("end_of_file", IntField(size=8)),
            ]
        )
        super().__init__()


class FileFullDirectoryInformation(Structure):
    """
    [MS-FSCC] 2.4.14 FileFullDirectoryInformation
    https://msdn.microsoft.com/en-us/library/cc232068.aspx
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_FULL_DIRECTORY_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("next_entry_offset", IntField(size=4)),
                ("file_index", IntField(size=4)),
                ("creation_time", DateTimeField(size=8)),
                ("last_access_time", DateTimeField(size=8)),
                ("last_write_time", DateTimeField(size=8)),
                ("change_time", DateTimeField(size=8)),
                ("end_of_file", IntField(size=8)),
                ("allocation_size", IntField(size=8)),
                (
                    "file_attributes",
                    FlagField(
                        size=4,
                        flag_type=FileAttributes,
                        flag_strict=False,
                    ),
                ),
                ("file_name_length", IntField(size=4, default=lambda s: len(s["file_name"]))),
                ("ea_size", IntField(size=4)),
                ("file_name", BytesField(size=lambda s: s["file_name_length"].get_value())),
            ]
        )
        super().__init__()


class FileFullEaInformation(Structure):
    """
    [MS-FSCC] 2.4.15 FileFullEaInformation
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/0eb94f48-6aac-41df-a878-79f4dcfd8989
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_FULL_EA_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("next_entry_offset", IntField(size=4)),
                ("flags", IntField(size=1)),
                (
                    "ea_name_length",
                    IntField(
                        size=1,
                        default=lambda s: len(s["ea_name"]),
                    ),
                ),
                (
                    "ea_value_length",
                    IntField(
                        size=2,
                        default=lambda s: len(s["ea_value"]),
                    ),
                ),
                (
                    "ea_name",
                    BytesField(
                        size=lambda s: s["ea_name_length"].get_value(),
                    ),
                ),
                # Not part of the spec byte EaName must be null padded so we add this field automatically.
                ("ea_name_padding", IntField(size=1)),
                (
                    "ea_value",
                    BytesField(
                        size=lambda s: s["ea_value_length"].get_value(),
                    ),
                ),
            ]
        )
        super().__init__()


class FileGetEaInformation(Structure):
    """
    [MS-FSCC] 2.4.15.1 FILE_GET_EA_INFORMATION
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/79dc1ea1-158c-4b24-b0e1-8c16c7e2af6b
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("next_entry_offset", IntField(size=4)),
                (
                    "ea_name_length",
                    IntField(
                        size=1,
                        default=lambda s: len(s["ea_name"]),
                    ),
                ),
                (
                    "ea_name",
                    BytesField(
                        size=lambda s: s["ea_name_length"].get_value(),
                    ),
                ),
                ("padding", IntField(size=1)),
            ]
        )
        super().__init__()


class FileIdBothDirectoryInformation(Structure):
    """
    [MS-FSCC] 2.4.17 FileIdBothDirectoryInformation
    https://msdn.microsoft.com/en-us/library/cc232070.aspx
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_ID_BOTH_DIRECTORY_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("next_entry_offset", IntField(size=4)),
                ("file_index", IntField(size=4)),
                ("creation_time", DateTimeField(size=8)),
                ("last_access_time", DateTimeField(size=8)),
                ("last_write_time", DateTimeField(size=8)),
                ("change_time", DateTimeField(size=8)),
                ("end_of_file", IntField(size=8)),
                ("allocation_size", IntField(size=8)),
                (
                    "file_attributes",
                    FlagField(
                        size=4,
                        flag_type=FileAttributes,
                        flag_strict=False,
                    ),
                ),
                ("file_name_length", IntField(size=4, default=lambda s: len(s["file_name"]))),
                ("ea_size", IntField(size=4)),
                ("short_name_length", IntField(size=1, default=lambda s: len(s["short_name"]))),
                ("reserved1", IntField(size=1)),
                ("short_name", BytesField(size=lambda s: s["short_name_length"].get_value())),
                (
                    "short_name_padding",
                    BytesField(
                        size=lambda s: 24 - len(s["short_name"]),
                        default=lambda s: b"\x00" * (24 - len(s["short_name"])),
                    ),
                ),
                ("reserved2", IntField(size=2)),
                ("file_id", IntField(size=8)),
                ("file_name", BytesField(size=lambda s: s["file_name_length"].get_value())),
            ]
        )
        super().__init__()


class FileIdFullDirectoryInformation(Structure):
    """
    [MS-FSCC] 2.4.18 FileIdFullDirectoryInformation
    https://msdn.microsoft.com/en-us/library/cc232071.aspx
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_ID_FULL_DIRECTORY_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("next_entry_offset", IntField(size=4)),
                ("file_index", IntField(size=4)),
                ("creation_time", DateTimeField(size=8)),
                ("last_access_time", DateTimeField(size=8)),
                ("last_write_time", DateTimeField(size=8)),
                ("change_time", DateTimeField(size=8)),
                ("end_of_file", IntField(size=8)),
                ("allocation_size", IntField(size=8)),
                (
                    "file_attributes",
                    FlagField(
                        size=4,
                        flag_type=FileAttributes,
                        flag_strict=False,
                    ),
                ),
                ("file_name_length", IntField(size=4, default=lambda s: len(s["file_name"]))),
                ("ea_size", IntField(size=4)),
                ("reserved", IntField(size=4)),
                ("file_id", IntField(size=8)),
                ("file_name", BytesField(size=lambda s: s["file_name_length"].get_value())),
            ]
        )
        super().__init__()


class FileInternalInformation(Structure):
    """
    [MS-FSSC] 2.4.20 FileInternalInformation
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/7d796611-2fa5-41ac-8178-b6fea3a017b3
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_INTERNAL_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("index_number", IntField(size=8)),
            ]
        )
        super().__init__()


class FileLinkInformation(Structure):
    """
    [MS-FSSC] 2.4.21 FileLinkInformation
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/69643dd3-b518-465d-bb0e-e2e9c5b7875e
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_LINK_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("replace_if_exists", BoolField()),
                (
                    "reserved",
                    BytesField(
                        size=7,
                        default=b"\x00" * 7,
                    ),
                ),
                ("root_directory", IntField(size=8)),
                (
                    "file_name_length",
                    IntField(
                        size=4,
                        default=lambda s: len(s["file_name"]),
                    ),
                ),
                (
                    "file_name",
                    TextField(
                        size=lambda s: s["file_name_length"].get_value(),
                        encoding="utf-16-le",
                    ),
                ),
            ]
        )
        super().__init__()


class FileModeInformation(Structure):
    """
    [MS-FSCC] 2.4.24 FileModeInformation
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/52df7798-8330-474b-ac31-9afe8075640c
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_MODE_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "mode",
                    FlagField(
                        size=4,
                        flag_type=ModeInformation,
                    ),
                ),
            ]
        )
        super().__init__()


class FileNamesInformation(Structure):
    """
    [MS-FSCC] 2.4.26 FileNamesInformation
    https://msdn.microsoft.com/en-us/library/cc232077.aspx
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_NAMES_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("next_entry_offset", IntField(size=4)),
                ("file_index", IntField(size=4)),
                ("file_name_length", IntField(size=4, default=lambda s: len(s["file_name"]))),
                ("file_name", BytesField(size=lambda s: s["file_name_length"].get_value())),
            ]
        )
        super().__init__()


class FilePositionInformation(Structure):
    """
    [MS-FSCC] 2.4.32 FilePositionInformation
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/e3ce4a39-327e-495c-99b6-6b61606b6f16
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_POSITION_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "current_byte_offset",
                    IntField(
                        size=8,
                        unsigned=False,
                    ),
                ),
            ]
        )
        super().__init__()


class FileRenameInformation(Structure):
    """
    [MS-FSCC] 2.4.34.2 FileRenameInformation For SMB2
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/52aa0b70-8094-4971-862d-79793f41e6a8
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_RENAME_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("replace_if_exists", BoolField()),
                (
                    "reserved",
                    BytesField(
                        size=7,
                        default=b"\x00" * 7,
                    ),
                ),
                ("root_directory", IntField(size=8)),
                (
                    "file_name_length",
                    IntField(
                        size=4,
                        default=lambda s: len(s["file_name"]),
                    ),
                ),
                (
                    "file_name",
                    TextField(
                        size=lambda s: s["file_name_length"].get_value(),
                    ),
                ),
            ]
        )
        super().__init__()


class FileStandardInformation(Structure):
    """
    [MS-FSCC] 2.4.38 FileStandardInformation
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/5afa7f66-619c-48f3-955f-68c4ece704ae
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_STANDARD_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "allocation_size",
                    IntField(
                        size=8,
                        unsigned=False,
                    ),
                ),
                (
                    "end_of_file",
                    IntField(
                        size=8,
                        unsigned=False,
                    ),
                ),
                ("number_of_links", IntField(size=4)),
                ("delete_pending", BoolField()),
                ("directory", BoolField()),
                ("reserved", IntField(size=2)),
            ]
        )
        super().__init__()


class FileStreamInformation(Structure):
    """
    [MS-FSCC] 2.4.40 FileStreamInformation
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/f8762be6-3ab9-411e-a7d6-5cc68f70c78d
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILE
    INFO_CLASS = FileInformationClass.FILE_STREAM_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("next_entry_offset", IntField(size=4)),
                (
                    "stream_name_length",
                    IntField(
                        size=4,
                        default=lambda s: len(s["stream_name"]),
                    ),
                ),
                (
                    "stream_size",
                    IntField(
                        size=8,
                        unsigned=False,
                    ),
                ),
                (
                    "stream_allocation_size",
                    IntField(
                        size=8,
                        unsigned=False,
                    ),
                ),
                (
                    "stream_name",
                    TextField(
                        encoding="utf-16-le",
                        size=lambda s: s["stream_name_length"].get_value(),
                    ),
                ),
            ]
        )
        super().__init__()


class FileFsFullSizeInformation(Structure):
    """
    [MS-FSCC] 2.5.4 FileFsFullSizeInformation
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/63768db7-9012-4209-8cca-00781e7322f5
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILESYSTEM
    INFO_CLASS = FileSystemInformationClass.FILE_FS_FULL_SIZE_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("total_allocation_units", IntField(size=8, unsigned=False)),
                ("caller_available_units", IntField(size=8, unsigned=False)),
                ("actual_available_units", IntField(size=8, unsigned=False)),
                ("sectors_per_unit", IntField(size=4)),
                ("bytes_per_sector", IntField(size=4)),
            ]
        )
        super().__init__()


class FileFsObjectIdInformation(Structure):
    """
    [MS-FSCC] 2.5.6 FileFsObjectIdInformation
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/dbf535ae-315a-4508-8bc5-84276ea106d4
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILESYSTEM
    INFO_CLASS = FileSystemInformationClass.FILE_FS_OBJECT_ID_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("object_id", UuidField()),
                (
                    "extended_info",
                    BytesField(
                        size=48,
                        default=b"\x00" * 48,
                    ),
                ),
            ]
        )
        super().__init__()


class FileFsVolumeInformation(Structure):
    """
    [MS-FSCC] 2.5.9 FileFsVolumeInformation
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/bf691378-c34e-4a13-976e-404ea1a87738
    """

    INFO_TYPE = InfoType.SMB2_0_INFO_FILESYSTEM
    INFO_CLASS = FileSystemInformationClass.FILE_FS_VOLUME_INFORMATION

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("volume_creation_time", DateTimeField()),
                ("volume_serial_number", IntField(size=4)),
                (
                    "volume_label_length",
                    IntField(
                        size=4,
                        default=lambda s: len(s["volume_label"]),
                    ),
                ),
                ("supports_objects", BoolField(default=False)),
                ("reserved", IntField(size=1)),
                (
                    "volume_label",
                    TextField(
                        size=lambda s: s["volume_label_length"].get_value(),
                        encoding="utf-16-le",
                    ),
                ),
            ]
        )
        super().__init__()



# Copyright: (c) 2019, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import binascii
import errno
import ntpath
import os
import socket
from collections import OrderedDict
from typing import Any

from smbprotocol import Dialects
from smbprotocol._text import to_bytes, to_text
from smbprotocol.header import NtStatus, SMB2HeaderResponse
from smbprotocol.reparse_point import SymbolicLinkReparseDataBuffer
from smbprotocol.structure import (
    BytesField,
    EnumField,
    IntField,
    ListField,
    Structure,
    StructureField,
)


class SMBException(Exception):
    # Generic SMB Exception with a message
    pass


class SMBAuthenticationError(SMBException):
    # Used for authentication specific errors
    pass


class SMBConnectionClosed(SMBException):
    # Used to denote the underlying TCP transport has been closed.
    pass


class SMBOSError(OSError, SMBException):
    """Wrapper for OSError with smbprotocol specific details.

    This is a wrapper for OSError for use in smbclient. It is designed to translate the raw SMBResponseException that
    smbprotocol raises into an error that is compatible with OSError.
    """

    def __init__(self, ntstatus, filename, filename2=None):
        self.ntstatus = ntstatus
        self.filename2 = os.fspath(filename2) if filename2 else None

        ntstatus_name = "STATUS_UNKNOWN"
        for name, val in vars(NtStatus).items():
            if ntstatus == val:
                ntstatus_name = name
                break

        error_details = {
            NtStatus.STATUS_OBJECT_NAME_NOT_FOUND: errno.ENOENT,
            NtStatus.STATUS_OBJECT_PATH_NOT_FOUND: errno.ENOENT,
            NtStatus.STATUS_NOT_FOUND: errno.ENOENT,
            NtStatus.STATUS_OBJECT_NAME_COLLISION: errno.EEXIST,
            NtStatus.STATUS_PRIVILEGE_NOT_HELD: (errno.EACCES, "Required privilege not held"),
            NtStatus.STATUS_SHARING_VIOLATION: (
                errno.EPERM,
                "The process cannot access the file because it is being used by another process",
            ),
            NtStatus.STATUS_NOT_A_REPARSE_POINT: (errno.EINVAL, "The file or directory is not a reparse point"),
            NtStatus.STATUS_FILE_IS_A_DIRECTORY: errno.EISDIR,
            NtStatus.STATUS_NOT_A_DIRECTORY: errno.ENOTDIR,
            NtStatus.STATUS_DIRECTORY_NOT_EMPTY: errno.ENOTEMPTY,
            NtStatus.STATUS_END_OF_FILE: getattr(errno, "ENODATA", 120),  # Not present on py2 for Windows.
        }.get(ntstatus, (0, f"Unknown NtStatus error returned '{ntstatus_name}'"))

        if not isinstance(error_details, tuple):
            error_details = (error_details, os.strerror(error_details))

        super().__init__(error_details[0], error_details[1], os.fspath(filename))

    def __str__(self):
        status = format(self.ntstatus, "x").zfill(8)
        msg = f"[Error {self.errno}] [NtStatus 0x{status}] {self.strerror}: '{self.filename}'"
        if self.filename2:
            msg += f" -> '{self.filename2}'"

        return msg


class SMBLinkRedirectionError(SMBException):
    @property
    def message(self):
        msg = (
            f"Encountered symlink at '{self.path}' that points to "
            f"'{self.target}' which cannot be redirected: {self.args[0]}"
        )
        return msg

    @property
    def path(self):
        return self.args[1]

    @property
    def target(self):
        return self.args[2]

    def __str__(self):
        return self.message


class SMBUnsupportedFeature(SMBException):
    @property
    def negotiated_dialect(self):
        return self.args[0]

    @property
    def required_dialect(self):
        return self.args[1]

    @property
    def feature_name(self):
        return self.args[2]

    @property
    def requires_newer(self):
        if len(self.args) > 3:
            return self.args[3]
        else:
            return None

    @property
    def message(self):
        if self.requires_newer is None:
            msg_suffix = ""
        elif self.requires_newer:
            msg_suffix = " or newer"
        else:
            msg_suffix = " or older"

        required_dialect = self._get_dialect_name(self.required_dialect)
        negotiated_dialect = self._get_dialect_name(self.negotiated_dialect)

        msg = (
            f"{self.feature_name} is not available on the negotiated dialect {negotiated_dialect}, "
            f"requires dialect {required_dialect}{msg_suffix}"
        )
        return msg

    def __str__(self):
        return self.message

    def _get_dialect_name(self, dialect):
        dialect_field = EnumField(enum_type=Dialects, enum_strict=False, size=2)
        dialect_field.set_value(dialect)
        return str(dialect_field)


class _SMBErrorRegistry(type):
    __registry = {}

    def __init__(cls, name, bases, attributes):
        super().__init__(name, bases, attributes)

        # Special case for the base SMBResponseException doesn't need to have _STATUS_CODE
        if cls.__module__ == _SMBErrorRegistry.__module__ and cls.__name__ == "SMBResponseException":
            return

        if not hasattr(cls, "_STATUS_CODE"):
            raise ValueError(f"{cls.__module__}.{cls.__name__} does not have the _STATUS_CODE class attribute set")

        cls.__registry[cls._STATUS_CODE] = cls

    def __call__(cls, header=None):
        # intercept construction of new SMBResponseException (and derived) instances
        if header:
            new_cls = cls.__registry.get(header["status"].get_value(), cls)

        else:
            header = SMB2HeaderResponse()
            header["status"] = cls._STATUS_CODE
            header["data"] = SMB2ErrorResponse()
            new_cls = cls

        # now use type.__call__ directly to actually construct instance
        return type.__call__(new_cls, header)


class SMBResponseException(SMBException, metaclass=_SMBErrorRegistry):
    """Base SMB response status exception.

    This is the base exception that is used when processing the status of an SMB response received from the server.
    When being called it may create one of the inherited types depending on whether the status is known or not. Each
    inherited class must implement _STATUS_CODE representing the NtStatus error it is for and may implement
    _BASE_MESSAGE to provide a better explanation of the error.
    """

    _BASE_MESSAGE = "Unknown error."

    def __init__(self, header: SMB2HeaderResponse):
        self.header = header

    @property
    def error_details(self) -> list[Any]:
        # list of error_details returned by the server, currently used in
        # the SMB 3.1.1 error response for certain situations
        error = SMB2ErrorResponse()
        error.unpack(self.header["data"].get_value())

        error_details = []
        for raw_error_data in error["error_data"].get_value():
            error_id = raw_error_data["error_id"].get_value()
            raw_data = raw_error_data["error_context_data"].get_value()
            if self.status == NtStatus.STATUS_STOPPED_ON_SYMLINK:
                error_data = SMB2SymbolicLinkErrorResponse()
                error_data.unpack(raw_data)

            elif (
                self.status == NtStatus.STATUS_BAD_NETWORK_NAME
                and error_id == ErrorContextId.SMB2_ERROR_ID_SHARE_REDIRECT
            ):
                error_data = SMB2ShareRedirectErrorContext()
                error_data.unpack(raw_data)

            else:
                # unknown context data so we just set it the raw bytes
                error_data = raw_data

            error_details.append(error_data)

        return error_details

    @property
    def message(self) -> str:
        error_details = []

        for detail in self.error_details:
            if isinstance(detail, SMB2SymbolicLinkErrorResponse):
                flag = str(detail["flags"])
                print_name = detail.get_print_name()
                sub_name = detail.get_substitute_name()
                error_details.append(f"Flag: {flag}, Print Name: {print_name}, Substitute Name: {sub_name}")

            elif isinstance(detail, SMB2ShareRedirectErrorContext):
                ip_addresses = []
                for ip_addr in detail["ip_addr_move_list"].get_value():
                    ip_addresses.append(ip_addr.get_ipaddress())

                resource_name = to_text(detail["resource_name"].get_value(), encoding="utf-16-le")
                addresses = "', '".join(ip_addresses)
                error_details.append(f"IP Addresses: '{addresses}', Resource Name: {resource_name}")

            else:
                # unknown error details in response, output raw bytes
                error_details.append("Raw: " + to_text(binascii.hexlify(detail)))

        status = format(self.status, "x").zfill(8)
        error_msg = f"{self._BASE_MESSAGE} {self.header['status']}: 0x{status}"
        if error_details:
            error_msg += " - " + (", ".join(error_details))

        return f"Received unexpected status from the server: {error_msg}"

    @property
    def status(self):
        return self.header["status"].get_value()

    def __str__(self):
        return self.message


class Unsuccessful(SMBResponseException):
    _BASE_MESSAGE = "The requested operation was unsuccessful."
    _STATUS_CODE = NtStatus.STATUS_UNSUCCESSFUL


class NetworkNameDelegated(SMBResponseException):
    _BASE_MESSAGE = "The network name was deleted."
    _STATUS_CODE = NtStatus.STATUS_NETWORK_NAME_DELETED


class StatusPending(SMBResponseException):
    _BASE_MESSAGE = "The operation that was requested is pending completion."
    _STATUS_CODE = NtStatus.STATUS_PENDING


class NotifyCleanup(SMBResponseException):
    _BASE_MESSAGE = (
        "Indicates that a notify change request has been completed due to closing the handle that made "
        "the notify change request."
    )
    _STATUS_CODE = NtStatus.STATUS_NOTIFY_CLEANUP


class NotifyEnumDir(SMBResponseException):
    _BASE_MESSAGE = (
        "Indicates that a notify change request is being completed and that the information is not "
        "being returned in the caller's buffer. The caller now needs to enumerate the files to find "
        "the changes."
    )
    _STATUS_CODE = NtStatus.STATUS_NOTIFY_ENUM_DIR


class BufferOverflow(SMBResponseException):
    _BASE_MESSAGE = "The data was too large to fit into the specified buffer."
    _STATUS_CODE = NtStatus.STATUS_BUFFER_OVERFLOW


class NoMoreFiles(SMBResponseException):
    _BASE_MESSAGE = "No more files were found which match the file specification."
    _STATUS_CODE = NtStatus.STATUS_NO_MORE_FILES


class EndOfFile(SMBResponseException):
    _BASE_MESSAGE = "The end-of-file marker has been reached. There is no valid data in the file beyond this marker."
    _STATUS_CODE = NtStatus.STATUS_END_OF_FILE


class InvalidEAName(SMBResponseException):
    _BASE_MESSAGE = "The specified extended attribute (EA) name contains at least one illegal character."
    _STATUS_CODE = NtStatus.STATUS_INVALID_EA_NAME


class EAListInconsistent(SMBResponseException):
    _BASE_MESSAGE = "The extended attribute (EA) list is inconsistent."
    _STATUS_CODE = NtStatus.STATUS_EA_LIST_INCONSISTENT


class StoppedOnSymlink(SMBResponseException):
    _BASE_MESSAGE = "The create operation stopped after reaching a symbolic link."
    _STATUS_CODE = NtStatus.STATUS_STOPPED_ON_SYMLINK


class InvalidInfoClass(SMBResponseException):
    _BASE_MESSAGE = "The specified information class is not a valid information class for the specified object."
    _STATUS_CODE = NtStatus.STATUS_INVALID_INFO_CLASS


class InfoLengthMismatch(SMBResponseException):
    _BASE_MESSAGE = (
        "The specified information record length does not match the length that is required for the "
        "specified information class."
    )
    _STATUS_CODE = NtStatus.STATUS_INFO_LENGTH_MISMATCH


class InvalidParameter(SMBResponseException):
    _BASE_MESSAGE = "An invalid parameter was passed to a service or function."
    _STATUS_CODE = NtStatus.STATUS_INVALID_PARAMETER


class NoSuchFile(SMBResponseException):
    _BASE_MESSAGE = "The file does not exist."
    _STATUS_CODE = NtStatus.STATUS_NO_SUCH_FILE


class InvalidDeviceRequest(SMBResponseException):
    _BASE_MESSAGE = "The specified request is not a valid operation for the target device."
    _STATUS_CODE = NtStatus.STATUS_INVALID_DEVICE_REQUEST


class MoreProcessingRequired(SMBResponseException):
    _BASE_MESSAGE = (
        "The specified I/O request packet (IRP) cannot be disposed of because the I/O operation is not complete."
    )
    _STATUS_CODE = NtStatus.STATUS_MORE_PROCESSING_REQUIRED


class AccessDenied(SMBResponseException):
    _BASE_MESSAGE = "A process has requested access to an object but has not been granted those access rights."
    _STATUS_CODE = NtStatus.STATUS_ACCESS_DENIED


class BufferTooSmall(SMBResponseException):
    _BASE_MESSAGE = "The buffer is too small to contain the entry. No information has been written to the buffer."
    _STATUS_CODE = NtStatus.STATUS_BUFFER_TOO_SMALL


class ObjectNameInvalid(SMBResponseException):
    _BASE_MESSAGE = "The object name is invalid."
    _STATUS_CODE = NtStatus.STATUS_OBJECT_NAME_INVALID


class ObjectNameNotFound(SMBResponseException):
    _BASE_MESSAGE = "The object name is not found."
    _STATUS_CODE = NtStatus.STATUS_OBJECT_NAME_NOT_FOUND


class ObjectNameCollision(SMBResponseException):
    _BASE_MESSAGE = "The object name already exists."
    _STATUS_CODE = NtStatus.STATUS_OBJECT_NAME_COLLISION


class ObjectPathInvalid(SMBResponseException):
    _BASE_MESSAGE = "The object path component was not a directory object."
    _STATUS_CODE = NtStatus.STATUS_OBJECT_PATH_INVALID


class ObjectPathNotFound(SMBResponseException):
    _BASE_MESSAGE = "The path does not exist."
    _STATUS_CODE = NtStatus.STATUS_OBJECT_PATH_NOT_FOUND


class ObjectPathSyntaxBad(SMBResponseException):
    _BASE_MESSAGE = "The object path component was not a directory object."
    _STATUS_CODE = NtStatus.STATUS_OBJECT_PATH_SYNTAX_BAD


class SharingViolation(SMBResponseException):
    _BASE_MESSAGE = "A file cannot be opened because the share access flags are incompatible."
    _STATUS_CODE = NtStatus.STATUS_SHARING_VIOLATION


class EASNotSupported(SMBResponseException):
    _BASE_MESSAGE = "An operation involving EAs failed because the file system does not support EAs."
    _STATUS_CODE = NtStatus.STATUS_EAS_NOT_SUPPORTED


class EATooLarge(SMBResponseException):
    _BASE_MESSAGE = "An EA operation failed because the EA set is too large."
    _STATUS_CODE = NtStatus.STATUS_EA_TOO_LARGE


class NonExistentEAEntry(SMBResponseException):
    _BASE_MESSAGE = "An EA operation failed because the name or EA index is invalid."
    _STATUS_CODE = NtStatus.STATUS_NONEXISTENT_EA_ENTRY


class NoEASOnFile(SMBResponseException):
    _BASE_MESSAGE = "The file for which EAs were requested has no EAs."
    _STATUS_CODE = NtStatus.STATUS_NO_EAS_ON_FILE


class EACorruptError(SMBResponseException):
    _BASE_MESSAGE = "The EA is corrupt and cannot be read."
    _STATUS_CODE = NtStatus.STATUS_EA_CORRUPT_ERROR


class DeletePending(SMBResponseException):
    _BASE_MESSAGE = "A non-close operation has been requested of a file object that has a delete pending."
    _STATUS_CODE = NtStatus.STATUS_DELETE_PENDING


class PrivilegeNotHeld(SMBResponseException):
    _BASE_MESSAGE = "A required privilege is not held by the client."
    _STATUS_CODE = NtStatus.STATUS_PRIVILEGE_NOT_HELD


class WrongPassword(SMBResponseException):
    _BASE_MESSAGE = "The specified password is not correct or the user is locked out."
    _STATUS_CODE = NtStatus.STATUS_WRONG_PASSWORD


class LogonFailure(SMBResponseException):
    _BASE_MESSAGE = (
        "The attempted logon is invalid. This is either due to a bad username or authentication information."
    )
    _STATUS_CODE = NtStatus.STATUS_LOGON_FAILURE


class PasswordExpired(SMBResponseException):
    _BASE_MESSAGE = "The user account password has expired."
    _STATUS_CODE = NtStatus.STATUS_PASSWORD_EXPIRED


class NoneMapped(SMBResponseException):
    _BASE_MESSAGE = "None of the information to be translated has been translated."
    _STATUS_CODE = NtStatus.STATUS_NONE_MAPPED


class InsufficientResources(SMBResponseException):
    _BASE_MESSAGE = "Insufficient system resources exist to complete the API."
    _STATUS_CODE = NtStatus.STATUS_INSUFFICIENT_RESOURCES


class PipeNotAvailable(SMBResponseException):
    _BASE_MESSAGE = "An instance of a named pipe cannot be found in the listening state."
    _STATUS_CODE = NtStatus.STATUS_PIPE_NOT_AVAILABLE


class PipeBusy(SMBResponseException):
    _BASE_MESSAGE = (
        "The specified pipe is set to complete operations and there are current I/O operations queued "
        "so that it cannot be changed to queue operations."
    )
    _STATUS_CODE = NtStatus.STATUS_PIPE_BUSY


class PipeClosing(SMBResponseException):
    _BASE_MESSAGE = "The specified named pipe is in the closing state."
    _STATUS_CODE = NtStatus.STATUS_PIPE_CLOSING


class IOTimeout(SMBResponseException):
    _BASE_MESSAGE = "The specified I/O operation was not completed before the time-out period expired."
    _STATUS_CODE = NtStatus.STATUS_IO_TIMEOUT


class PipeDisconnected(SMBResponseException):
    _BASE_MESSAGE = "The specified named pipe is in the disconnected state."
    _STATUS_CODE = NtStatus.STATUS_PIPE_DISCONNECTED


class FileIsADirectory(SMBResponseException):
    _BASE_MESSAGE = (
        "The file that was specified as a target is a directory, and the caller specified that it could "
        "be anything but a directory."
    )
    _STATUS_CODE = NtStatus.STATUS_FILE_IS_A_DIRECTORY


class NotSupported(SMBResponseException):
    _BASE_MESSAGE = "The request is not supported."
    _STATUS_CODE = NtStatus.STATUS_NOT_SUPPORTED


class BadNetworkName(SMBResponseException):
    _BASE_MESSAGE = "The specified share name cannot be found on the remote server."
    _STATUS_CODE = NtStatus.STATUS_BAD_NETWORK_NAME


class RequestNotAccepted(SMBResponseException):
    _BASE_MESSAGE = (
        "No more connections can be made to this remote computer at this time because the computer has "
        "already accepted the maximum number of connections."
    )
    _STATUS_CODE = NtStatus.STATUS_REQUEST_NOT_ACCEPTED


class PipeEmpty(SMBResponseException):
    _BASE_MESSAGE = "Used to indicate that a read operation was done on an empty pipe."
    _STATUS_CODE = NtStatus.STATUS_PIPE_EMPTY


class InternalError(SMBResponseException):
    _BASE_MESSAGE = "An internal error occurred."
    _STATUS_CODE = NtStatus.STATUS_INTERNAL_ERROR


class DirectoryNotEmpty(SMBResponseException):
    _BASE_MESSAGE = "Indicates that the directory trying to be deleted is not empty."
    _STATUS_CODE = NtStatus.STATUS_DIRECTORY_NOT_EMPTY


class StatusNotADirectory(SMBResponseException):
    _BASE_MESSAGE = "A requested opened file is not a directory."
    _STATUS_CODE = NtStatus.STATUS_NOT_A_DIRECTORY


class Cancelled(SMBResponseException):
    _BASE_MESSAGE = "The I/O request was canceled."
    _STATUS_CODE = NtStatus.STATUS_CANCELLED


class CannotDelete(SMBResponseException):
    _BASE_MESSAGE = "An attempt has been made to remove a file or directory that cannot be deleted."
    _STATUS_CODE = NtStatus.STATUS_CANNOT_DELETE


class FileClosed(SMBResponseException):
    _BASE_MESSAGE = (
        "An I/O request other than close and several other special case operations was attempted using "
        "a file object that had already been closed."
    )
    _STATUS_CODE = NtStatus.STATUS_FILE_CLOSED


class PipeBroken(SMBResponseException):
    _BASE_MESSAGE = "The pipe operation has failed because the other end of the pipe has been closed."
    _STATUS_CODE = NtStatus.STATUS_PIPE_BROKEN


class FSDriverRequired(SMBResponseException):
    _BASE_MESSAGE = (
        "A volume has been accessed for which a file system driver is required that has not yet been loaded."
    )
    _STATUS_CODE = NtStatus.STATUS_FS_DRIVER_REQUIRED


class UserSessionDeleted(SMBResponseException):
    _BASE_MESSAGE = "The remote user session has been deleted."
    _STATUS_CODE = NtStatus.STATUS_USER_SESSION_DELETED


class InsuffServerResources(SMBResponseException):
    _BASE_MESSAGE = "Insufficient server resources exist to complete the request."
    _STATUS_CODE = NtStatus.STATUS_INSUFF_SERVER_RESOURCES


class NotFound(SMBResponseException):
    _BASE_MESSAGE = "The object was not found."
    _STATUS_CODE = NtStatus.STATUS_NOT_FOUND


class PathNotCovered(SMBResponseException):
    _BASE_MESSAGE = "The contacted server does not support the indicated part of the DFS namespace."
    _STATUS_CODE = NtStatus.STATUS_PATH_NOT_COVERED


class DfsUnavailable(SMBResponseException):
    _BASE_MESSAGE = "DFS is unavailable on the contacted server."
    _STATUS_CODE = NtStatus.STATUS_DFS_UNAVAILABLE


class NotAReparsePoint(SMBResponseException):
    _BASE_MESSAGE = "The NTFS file or directory is not a reparse point."
    _STATUS_CODE = NtStatus.STATUS_NOT_A_REPARSE_POINT


class ServerUnavailable(SMBResponseException):
    _BASE_MESSAGE = "The file server is temporarily unavailable."
    _STATUS_CODE = NtStatus.STATUS_SERVER_UNAVAILABLE


class DiskFull(SMBResponseException):
    _BASE_MESSAGE = "There is not enough space on the disk."
    _STATUS_CODE = NtStatus.STATUS_DISK_FULL


class ErrorContextId:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.2.1 SMB2 Error Context Response ErrorId
    An identifier for the error context, it MUST be set to one of the following
    values.
    """

    SMB2_ERROR_ID_DEFAULT = 0x00000000
    SMB2_ERROR_ID_SHARE_REDIRECT = 0x53526472


class SymbolicLinkErrorFlags:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.2.2.1 Symbolic Link Error Response Flags
    Specifies whether the substitute name is an absolute target path or a path
    name relative to the directory containing the symbolic link
    """

    SYMLINK_FLAG_ABSOLUTE = 0x00000000
    SYMLINK_FLAG_RELATIVE = 0x00000001


class IpAddrType:
    """
    [MS-SM2] v53.0 2017-09-15

    2.2.2.2.2.1 MOVE_DST_IPADDR structure Type
    Indicates the type of the destionation IP address.
    """

    MOVE_DST_IPADDR_V4 = 0x00000001
    MOVE_DST_IPADDR_V6 = 0x00000002


class SMB2ErrorResponse(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.2 SMB2 Error Response
    The SMB2 Error Response packet is sent by the server to respond to a
    request that has failed or encountered an error. This is only used in the
    SMB 3.1.1 dialect and this code won't decode values based on older versions
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "structure_size",
                    IntField(
                        size=2,
                        default=9,
                    ),
                ),
                (
                    "error_context_count",
                    IntField(
                        size=1,
                        default=lambda s: len(s["error_data"].get_value()),
                    ),
                ),
                ("reserved", IntField(size=1)),
                (
                    "byte_count",
                    IntField(
                        size=4,
                        default=lambda s: len(s["error_data"]),
                    ),
                ),
                (
                    "error_data",
                    ListField(
                        size=lambda s: s["byte_count"].get_value(),
                        list_count=lambda s: s["error_context_count"].get_value(),
                        list_type=StructureField(structure_type=SMB2ErrorContextResponse),
                        unpack_func=lambda s, d: self._error_data_value(s, d),
                    ),
                ),
            ]
        )
        super().__init__()

    def _error_data_value(self, structure, data):
        context_responses = []

        while len(data) > 0:
            response = SMB2ErrorContextResponse()
            if structure["error_context_count"].get_value() > 0:
                # Working with SMB 3.1.1+ where the errors are already in an SMB2ErrorContextReponse packet, unpack the
                # data as usual
                data = response.unpack(data)
            else:
                # Working with an older SMB dialect where the response is set directly in the error_data field, need to
                # manually craft the SMB2ErrorContextResponse with the data returned.
                response["error_context_data"] = data
                data = b""

            context_responses.append(response)

        return context_responses


class SMB2ErrorContextResponse(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.2.1 SMB2 ERROR Context Response
    For the SMB dialect 3.1.1, the server formats the error data as an array of
    SMB2 Error Context structures in the SMB2ErrorResponse message.

    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "error_data_length",
                    IntField(
                        size=4,
                        default=lambda s: len(s["error_context_data"]),
                    ),
                ),
                (
                    "error_id",
                    EnumField(
                        size=4,
                        default=ErrorContextId.SMB2_ERROR_ID_DEFAULT,
                        enum_type=ErrorContextId,
                    ),
                ),
                (
                    "error_context_data",
                    BytesField(
                        size=lambda s: s["error_data_length"].get_value(),
                    ),
                ),
            ]
        )
        super().__init__()


class SMB2SymbolicLinkErrorResponse(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.2.2.1 Symbolic Link Error Response
    The Symbolic Link Error Response is used to indicate that a symbolic link
    was encountered on the create. It describes the target path that the client
    MUST use if it requires to follow the symbolic link.
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("symlink_length", IntField(size=4, default=lambda s: len(s) - 4)),
                ("symlink_error_tag", BytesField(size=4, default=b"\x53\x59\x4d\x4c")),
                ("reparse_tag", BytesField(size=4, default=b"\x0c\x00\x00\xa0")),
                ("reparse_data_length", IntField(size=2, default=lambda s: len(s["path_buffer"]) + 12)),
                # the len in utf-16-le bytes of the path beyond the substitute name
                # of the original target, e.g. \\server\share\symlink\file.txt
                # would be length of \file.txt in utf-16-le form, this is used by
                # the client to find out what part of the original path to append
                # to the substitute name returned by the server.
                ("unparsed_path_length", IntField(size=2)),
                ("substitute_name_offset", IntField(size=2)),
                ("substitute_name_length", IntField(size=2)),
                ("print_name_offset", IntField(size=2)),
                ("print_name_length", IntField(size=2)),
                (
                    "flags",
                    EnumField(
                        size=2,
                        enum_type=SymbolicLinkErrorFlags,
                    ),
                ),
                # Not in the spec but Windows seems to add \x00\x80 to the end of flags which breaks our parsing. Cannot
                # seem to figure out why but this just ignored that field.
                ("reserved", IntField(size=2)),
                # use the get/set_name functions to get/set these values as they
                # also (d)encode the text and set the length and offset accordingly
                ("path_buffer", BytesField(size=lambda s: self._get_name_length(s, True))),
            ]
        )
        super().__init__()

    def _get_name_length(self, structure, first):
        print_name_len = structure["print_name_length"].get_value()
        sub_name_len = structure["substitute_name_length"].get_value()
        return print_name_len + sub_name_len

    def set_name(self, print_name, substitute_name):
        """
        Set's the path_buffer and print/substitute name length of the message
        with the values passed in. These values should be a string and not a
        byte string as it is encoded in this function.

        :param print_name: The print name string to set
        :param substitute_name: The substitute name string to set
        """
        # Ensure that the to_bytes input is an actual text string for py2 compat with native strings.
        print_bytes = to_bytes(to_text(print_name), encoding="utf-16-le")
        sub_bytes = to_bytes(to_text(substitute_name), encoding="utf-16-le")
        path_buffer = print_bytes + sub_bytes

        self["print_name_offset"].set_value(0)
        self["print_name_length"].set_value(len(print_bytes))
        self["substitute_name_offset"].set_value(len(print_bytes))
        self["substitute_name_length"].set_value(len(sub_bytes))
        self["path_buffer"].set_value(path_buffer)

    def get_print_name(self):
        offset = self["print_name_offset"].get_value()
        length = self["print_name_length"].get_value()
        name_bytes = self["path_buffer"].get_value()[offset : offset + length]
        return to_text(name_bytes, encoding="utf-16-le")

    def get_substitute_name(self):
        offset = self["substitute_name_offset"].get_value()
        length = self["substitute_name_length"].get_value()
        name_bytes = self["path_buffer"].get_value()[offset : offset + length]
        return to_text(name_bytes, encoding="utf-16-le")

    def resolve_path(self, link_path):
        """
        [MS-SMB2] 2.2.2.2.1.1 Handling the Symbolic Link Error Response
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/a8da655c-8b0b-415a-b726-16dc33fa5827

        Attempts to resolve the link target path. Will fail if the link is pointing to a local path or a UNC path
        on another host or share.

        :param link_path: The original path to the symbolic link to resolve relative paths from.
        :return: The resolved link target path.
        """
        substitute_name = self.get_substitute_name()
        print_name = self.get_print_name()
        unparsed_path_length = self["unparsed_path_length"].get_value()

        b_link_path = to_bytes(to_text(link_path), encoding="utf-16-le")
        unparsed_idx = len(b_link_path) - unparsed_path_length
        base_link_path = to_text(b_link_path[:unparsed_idx], encoding="utf-16-le")
        unparsed_path = to_text(b_link_path[unparsed_idx:], encoding="utf-16-le")

        # Use the common code in SymbolicLinkReparseDataBuffer() to resolve the link target.
        symlink_buffer = SymbolicLinkReparseDataBuffer()
        symlink_buffer["flags"] = self["flags"].get_value()
        symlink_buffer.set_name(substitute_name, print_name)
        target_path = symlink_buffer.resolve_link(base_link_path) + unparsed_path

        if not target_path.startswith("\\\\"):
            raise SMBLinkRedirectionError(
                "Cannot resolve link targets that point to a local path", link_path, print_name
            )

        link_share = ntpath.splitdrive(link_path)[0]
        target_share = ntpath.splitdrive(target_path)[0]
        if link_share != target_share:
            raise SMBLinkRedirectionError(
                "Cannot resolve link targets that point to a different host/share", link_path, print_name
            )

        return target_path


class SMB2ShareRedirectErrorContext(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.2.2.2 Share Redirect Error Context Response
    Response to a Tree Connect with the
    SMB2_TREE_CONNECT_FLAG_REDIRECT_TO_OWNER flag set.
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("structure_size", IntField(size=4, default=lambda s: len(s))),
                ("notification_type", IntField(size=4, default=3)),
                ("resource_name_offset", IntField(size=4, default=lambda s: self._resource_name_offset(s))),
                ("resource_name_length", IntField(size=4, default=lambda s: len(s["resource_name"]))),
                ("flags", IntField(size=2, default=0)),
                ("target_type", IntField(size=2, default=0)),
                ("ip_addr_count", IntField(size=4, default=lambda s: len(s["ip_addr_move_list"].get_value()))),
                (
                    "ip_addr_move_list",
                    ListField(
                        size=lambda s: s["ip_addr_count"].get_value() * 24,
                        list_count=lambda s: s["ip_addr_count"].get_value(),
                        list_type=StructureField(size=24, structure_type=SMB2MoveDstIpAddrStructure),
                    ),
                ),
                ("resource_name", BytesField(size=lambda s: s["resource_name_length"].get_value())),
            ]
        )
        super().__init__()

    def _resource_name_offset(self, structure):
        min_structure_size = 24
        addr_list_size = len(structure["ip_addr_move_list"])
        return min_structure_size + addr_list_size


class SMB2MoveDstIpAddrStructure(Structure):
    """
    [MS-SMB2] c53.0 2017-09-15

    2.2.2.2.2.1 MOVE_DST_IPADDR structure
    Used to indicate the destination IP address.
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("type", EnumField(size=4, enum_type=IpAddrType)),
                ("reserved", IntField(size=4)),
                ("ip_address", BytesField(size=lambda s: self._ip_address_size(s))),
                (
                    "reserved2",
                    BytesField(
                        size=lambda s: self._reserved2_size(s), default=lambda s: b"\x00" * self._reserved2_size(s)
                    ),
                ),
            ]
        )
        super().__init__()

    def _ip_address_size(self, structure):
        if structure["type"].get_value() == IpAddrType.MOVE_DST_IPADDR_V4:
            return 4
        else:
            return 16

    def _reserved2_size(self, structure):
        if structure["type"].get_value() == IpAddrType.MOVE_DST_IPADDR_V4:
            return 12
        else:
            return 0

    def get_ipaddress(self):
        # get's the IP address in a human readable format
        ip_address = self["ip_address"].get_value()
        if self["type"].get_value() == IpAddrType.MOVE_DST_IPADDR_V4:
            return socket.inet_ntoa(ip_address)
        else:
            addr = binascii.hexlify(ip_address).decode("utf-8")
            return ":".join([addr[i : i + 4] for i in range(0, len(addr), 4)])

    def set_ipaddress(self, address):
        # set's the IP address from a human readable format, for IPv6, this
        # needs to be the full IPv6 address
        if self["type"].get_value() == IpAddrType.MOVE_DST_IPADDR_V4:
            self["ip_address"].set_value(socket.inet_aton(address))
        else:
            addr = address.replace(":", "")
            if len(addr) != 32:
                raise ValueError("When setting an IPv6 address, it must be in the full form without concatenation")
            self["ip_address"].set_value(binascii.unhexlify(addr))



# Copyright: (c) 2019, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import logging
from collections import OrderedDict

from smbprotocol import Dialects
from smbprotocol.exceptions import (
    FileClosed,
    InvalidDeviceRequest,
    NotSupported,
    SMBException,
)
from smbprotocol.header import Commands
from smbprotocol.ioctl import (
    CtlCode,
    IOCTLFlags,
    SMB2IOCTLRequest,
    SMB2IOCTLResponse,
    SMB2ValidateNegotiateInfoRequest,
    SMB2ValidateNegotiateInfoResponse,
)
from smbprotocol.structure import BytesField, EnumField, FlagField, IntField, Structure

log = logging.getLogger(__name__)


class TreeFlags:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.9 SMB2 TREE_CONNECT Response Flags
    Flags used in SMB 3.1.1  to indicate how to process the operation.
    """

    SMB2_TREE_CONNECT_FLAG_CLUSTER_RECONNECT = 0x0004
    SMB2_TREE_CONNECT_FLAG_REDIRECT_TO_OWNER = 0x0002
    SMB2_TREE_CONNECT_FLAG_EXTENSION_PRESENT = 0x0001


class ShareType:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.10 SMB2 TREE_CONNECT Response Capabilities
    The type of share being accessed
    """

    SMB2_SHARE_TYPE_DISK = 0x01
    SMB2_SHARE_TYPE_PIPE = 0x02
    SMB2_SHARE_TYPE_PRINT = 0x03


class ShareFlags:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.10 SMB2 TREE_CONNECT Response Capabilities
    Properties for the share
    """

    SMB2_SHAREFLAG_MANUAL_CACHING = 0x00000000
    SMB2_SHAREFLAG_AUTO_CACHING = 0x00000010
    SMB2_SHAREFLAG_VDO_CACHING = 0x00000020
    SMB2_SHAREFLAG_NO_CACHING = 0x00000030
    SMB2_SHAREFLAG_DFS = 0x00000001
    SMB2_SHAREFLAG_DFS_ROOT = 0x00000002
    SMB2_SHAREFLAG_RESTRICT_EXCLUSIVE_OPENS = 0x00000100
    SMB2_SHAREFLAG_FORCE_SHARED_DELETE = 0x00000200
    SMB2_SHAREFLAG_ALLOW_NAMESPACE_CACHING = 0x00000400
    SMB2_SHAREFLAG_ACCESS_BASED_DIRECTORY_ENUM = 0x00000800
    SMB2_SHAREFLAG_FORCE_LEVELII_OPLOCK = 0x00001000
    SMB2_SHAREFLAG_ENABLE_HASH_V1 = 0x00002000
    SMB2_SHAREFLAG_ENABLE_HASH_V2 = 0x00004000
    SMB2_SHAREFLAG_ENCRYPT_DATA = 0x00008000
    SMB2_SHAREFLAG_IDENTITY_REMOTING = 0x00040000


class ShareCapabilities:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.10 SMB2 TREE_CONNECT Response Capabilities
    Indicates various capabilities for a share
    """

    SMB2_SHARE_CAP_DFS = 0x00000008
    SMB2_SHARE_CAP_CONTINUOUS_AVAILABILITY = 0x00000010
    SMB2_SHARE_CAP_SCALEOUT = 0x00000020
    SMB2_SHARE_CAP_CLUSTER = 0x00000040
    SMB2_SHARE_CAP_ASYMMETRIC = 0x00000080
    SMB2_SHARE_CAP_REDIRECT_TO_OWNER = 0x00000100


class SMB2TreeConnectRequest(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.9 SMB2 TREE_CONNECT Request
    Sent by the client to request access to a particular share on the server
    """

    COMMAND = Commands.SMB2_TREE_CONNECT

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("structure_size", IntField(size=2, default=9)),
                (
                    "flags",
                    FlagField(
                        size=2,
                        flag_type=TreeFlags,
                    ),
                ),
                (
                    "path_offset",
                    IntField(
                        size=2,
                        default=64 + 8,
                    ),
                ),
                (
                    "path_length",
                    IntField(
                        size=2,
                        default=lambda s: len(s["buffer"]),
                    ),
                ),
                ("buffer", BytesField(size=lambda s: s["path_length"].get_value())),
            ]
        )
        super().__init__()


class SMB2TreeConnectResponse(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.10 SMB2 TREE_CONNECT Response
    Sent by the server when an SMB2 TREE_CONNECT request is processed
    successfully.
    """

    COMMAND = Commands.SMB2_TREE_CONNECT

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("structure_size", IntField(size=2, default=16)),
                (
                    "share_type",
                    EnumField(
                        size=1,
                        enum_type=ShareType,
                    ),
                ),
                ("reserved", IntField(size=1)),
                (
                    "share_flags",
                    FlagField(
                        size=4,
                        flag_type=ShareFlags,
                    ),
                ),
                (
                    "capabilities",
                    FlagField(
                        size=4,
                        flag_type=ShareCapabilities,
                    ),
                ),
                ("maximal_access", IntField(size=4)),
            ]
        )
        super().__init__()


class SMB2TreeDisconnect(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.11/12 SMB2 TREE_DISCONNECT Request and Response
    Sent by the client to request that the tree connect specific by tree_id in
    the header is disconnected.
    """

    COMMAND = Commands.SMB2_TREE_DISCONNECT

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "structure_size",
                    IntField(
                        size=2,
                        default=4,
                    ),
                ),
                ("reserved", IntField(size=2)),
            ]
        )
        super().__init__()


class TreeConnect:
    def __init__(self, session, share_name):
        """
        [MS-SMB2] v53.0 2017-09-15

        3.2.1.4 Per Tree Connect
        Attributes per Tree Connect (share connections)

        :param session: The Session to connect to the tree with.
        :param share_name: The name of the share, including the server name.
        """
        self._connected = False
        self.open_table = {}

        self.share_name = share_name
        self.tree_connect_id = None
        self.session = session
        self.is_dfs_share = None

        # SMB 3.x+
        self.is_ca_share = None
        self.encrypt_data = None
        self.is_scaleout_share = None

    def connect(self, require_secure_negotiate=True):
        """
        Connect to the share.

        :param require_secure_negotiate: For Dialects 3.0 and 3.0.2, will
            verify the negotiation parameters with the server to prevent
            SMB downgrade attacks
        """
        log.info(f"Session: {self.session.username} - Creating connection to share {self.share_name}")
        utf_share_name = self.share_name.encode("utf-16-le")
        connect = SMB2TreeConnectRequest()
        connect["buffer"] = utf_share_name

        log.info("Session: %s - Sending Tree Connect message", self.session.username)
        log.debug(connect)
        request = self.session.connection.send(connect, sid=self.session.session_id)

        log.info("Session: %s - Receiving Tree Connect response", self.session.username)
        response = self.session.connection.receive(request)
        tree_response = SMB2TreeConnectResponse()
        tree_response.unpack(response["data"].get_value())
        log.debug(tree_response)

        # https://msdn.microsoft.com/en-us/library/cc246687.aspx
        self.tree_connect_id = response["tree_id"].get_value()
        log.info("Session: %s - Created tree connection with ID %d", self.session.username, self.tree_connect_id)
        self._connected = True
        self.session.tree_connect_table[self.tree_connect_id] = self

        capabilities = tree_response["capabilities"]
        self.is_dfs_share = capabilities.has_flag(ShareCapabilities.SMB2_SHARE_CAP_DFS)
        self.is_ca_share = capabilities.has_flag(ShareCapabilities.SMB2_SHARE_CAP_CONTINUOUS_AVAILABILITY)

        dialect = self.session.connection.dialect
        if dialect >= Dialects.SMB_3_0_0 and self.session.connection.supports_encryption:
            self.encrypt_data = tree_response["share_flags"].has_flag(ShareFlags.SMB2_SHAREFLAG_ENCRYPT_DATA)

            self.is_scaleout_share = capabilities.has_flag(ShareCapabilities.SMB2_SHARE_CAP_SCALEOUT)

        if require_secure_negotiate:
            self._verify_dialect_negotiate()

    def disconnect(self):
        """
        Disconnects the tree connection.
        """
        if not self._connected:
            return

        log.info(f"Session: {self.session.username}, Tree: {self.share_name} - Disconnecting from Tree Connect")

        req = SMB2TreeDisconnect()
        log.info(f"Session: {self.session.username}, Tree: {self.share_name} - Sending Tree Disconnect message")
        log.debug(req)
        request = self.session.connection.send(req, sid=self.session.session_id, tid=self.tree_connect_id)

        log.info(f"Session: {self.session.username}, Tree: {self.share_name} - Receiving Tree Disconnect response")
        res = self.session.connection.receive(request)
        res_disconnect = SMB2TreeDisconnect()
        res_disconnect.unpack(res["data"].get_value())
        log.debug(res_disconnect)
        self._connected = False
        del self.session.tree_connect_table[self.tree_connect_id]

    def _verify_dialect_negotiate(self):
        log_header = f"Session: {self.session.username}, Tree: {self.share_name}"
        log.info("%s - Running secure negotiate process", log_header)

        if not self.session.signing_key:
            # This will only happen if we authenticated with the guest or anonymous user.
            raise SMBException(
                "Cannot verify negotiate information without a session signing key. Authenticate with "
                "a non-guest or anonymous account or set require_secure_negotiate=False to disable the "
                "negotiation info verification checks."
            )

        dialect = self.session.connection.dialect
        if dialect >= Dialects.SMB_3_1_1:
            # SMB 3.1.1+ uses the negotiation info to generate the signing key so doesn't need this extra exchange.
            return

        ioctl_request = SMB2IOCTLRequest()
        ioctl_request["ctl_code"] = CtlCode.FSCTL_VALIDATE_NEGOTIATE_INFO
        ioctl_request["file_id"] = b"\xff" * 16

        val_neg = SMB2ValidateNegotiateInfoRequest()
        val_neg["capabilities"] = self.session.connection.client_capabilities
        val_neg["guid"] = self.session.connection.client_guid
        val_neg["security_mode"] = self.session.connection.client_security_mode
        val_neg["dialects"] = self.session.connection.negotiated_dialects

        ioctl_request["buffer"] = val_neg
        ioctl_request["max_output_response"] = len(val_neg)
        ioctl_request["flags"] = IOCTLFlags.SMB2_0_IOCTL_IS_FSCTL
        log.info("%s - Sending Secure Negotiate Validation message", log_header)
        log.debug(ioctl_request)
        request = self.session.connection.send(
            ioctl_request, sid=self.session.session_id, tid=self.tree_connect_id, force_signature=True
        )

        log.info("%s - Receiving secure negotiation response", log_header)
        try:
            response = self.session.connection.receive(request)

        except (FileClosed, InvalidDeviceRequest, NotSupported) as e:
            # https://docs.microsoft.com/en-us/archive/blogs/openspecification/smb3-secure-dialect-negotiation
            # Older dialects may respond with these exceptions, this is expected and we only want to fail if
            # they are not signed. Check that header signature was signed, fail if it wasn't. The signature, if
            # present, would have been verified when the connection received the data.
            if e.header["signature"].get_value() == b"\x00" * 16:
                raise

            return

        # If we received an actual response we want to validate the info provided matches with what was negotiated.
        ioctl_resp = SMB2IOCTLResponse()
        ioctl_resp.unpack(response["data"].get_value())
        log.debug(ioctl_resp)

        log.info("%s - Unpacking secure negotiate response info", log_header)
        val_resp = SMB2ValidateNegotiateInfoResponse()
        val_resp.unpack(ioctl_resp["buffer"].get_value())
        log.debug(val_resp)

        self._verify(
            "server capabilities",
            val_resp["capabilities"].get_value(),
            self.session.connection.server_capabilities.get_value(),
        )
        self._verify("server guid", val_resp["guid"].get_value(), self.session.connection.server_guid)
        self._verify(
            "server security mode", val_resp["security_mode"].get_value(), self.session.connection.server_security_mode
        )
        self._verify("server dialect", val_resp["dialect"].get_value(), self.session.connection.dialect)
        log.info("Session: %d, Tree: %d - Secure negotiate complete", self.session.session_id, self.tree_connect_id)

    def _verify(self, check, actual, expected):
        if actual != expected:
            log_header = f"Session: {self.session.session_id}, Tree: {self.tree_connect_id}"
            raise SMBException(
                f"{log_header} - Secure negotiate failed to verify {check}, Actual: {actual}, Expected: {expected}",
            )



# Copyright: (c) 2019, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import logging
from collections import OrderedDict

from smbprotocol import MAX_PAYLOAD_SIZE, Dialects
from smbprotocol.create_contexts import SMB2CreateContextRequest
from smbprotocol.exceptions import FileClosed, SMBException, SMBUnsupportedFeature
from smbprotocol.file_info import (  # While this shouldn't ever be removed, we need to keep this imported so we stay backwards compat. These were; originally defined here.
    FileAttributes,
    FileBothDirectoryInformation,
    FileDirectoryInformation,
    FileFullDirectoryInformation,
    FileFullEaInformation,
    FileIdBothDirectoryInformation,
    FileIdFullDirectoryInformation,
    FileInformationClass,
    FileNamesInformation,
    FileStreamInformation,
    InfoType,
)
from smbprotocol.header import Commands
from smbprotocol.structure import (
    BytesField,
    DateTimeField,
    EnumField,
    FlagField,
    IntField,
    ListField,
    Structure,
    StructureField,
)

log = logging.getLogger(__name__)


class RequestedOplockLevel:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.31 SMB2 CREATE Request RequestedOplockLevel
    The requested oplock level used when creating/accessing a file.
    """

    SMB2_OPLOCK_LEVEL_NONE = 0x00
    SMB2_OPLOCK_LEVEL_II = 0x01
    SMB2_OPLOCK_LEVEL_EXCLUSIVE = 0x08
    SMB2_OPLOCK_LEVEL_BATCH = 0x09
    SMB2_OPLOCK_LEVEL_LEASE = 0xFF


class ImpersonationLevel:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.31 SMB2 CREATE Request ImpersonationLevel
    The impersonation level requested by the application in a create request.
    """

    Anonymous = 0x0
    Identification = 0x1
    Impersonation = 0x2
    Delegate = 0x3


class ShareAccess:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.31 SMB2 CREATE Request ShareAccess
    The sharing mode for the open
    """

    FILE_SHARE_READ = 0x1
    FILE_SHARE_WRITE = 0x2
    FILE_SHARE_DELETE = 0x4


class CreateDisposition:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.31 SMB2 CREATE Request CreateDisposition
    Defines the action the server must take if the file that is specific
    already exists.
    """

    FILE_SUPERSEDE = 0x0
    FILE_OPEN = 0x1
    FILE_CREATE = 0x2
    FILE_OPEN_IF = 0x3
    FILE_OVERWRITE = 0x4
    FILE_OVERWRITE_IF = 0x5


class CreateOptions:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.31 SMB2 CREATE Request CreateOptions
    Specifies the options to be applied when creating or opening the file
    """

    FILE_DIRECTORY_FILE = 0x00000001
    FILE_WRITE_THROUGH = 0x00000002
    FILE_SEQUENTIAL_ONLY = 0x00000004
    FILE_NO_INTERMEDIATE_BUFFERING = 0x00000008
    FILE_SYNCHRONOUS_IO_ALERT = 0x00000010
    FILE_SYNCHRONOUS_IO_NONALERT = 0x00000020
    FILE_NON_DIRECTORY_FILE = 0x00000040
    FILE_COMPLETE_IF_OPLOCKED = 0x00000100
    FILE_NO_EA_KNOWLEDGE = 0x00000200
    FILE_RANDOM_ACCESS = 0x00000800
    FILE_DELETE_ON_CLOSE = 0x00001000
    FILE_OPEN_BY_FILE_ID = 0x00002000
    FILE_OPEN_FOR_BACKUP_INTENT = 0x00004000
    FILE_NO_COMPRESSION = 0x00008000
    FILE_OPEN_REMOTE_INSTANCE = 0x00000400
    FILE_OPEN_REQUIRING_OPLOCK = 0x00010000
    FILE_DISALLOW_EXCLUSIVE = 0x00020000
    FILE_RESERVE_OPFILTER = 0x00100000
    FILE_OPEN_REPARSE_POINT = 0x00200000
    FILE_OPEN_NO_RECALL = 0x00400000
    FILE_OPEN_FOR_FREE_SPACE_QUERY = 0x00800000


class FilePipePrinterAccessMask:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.13.1.1 File_Pipe_Printer_Access_Mask
    Access Mask flag values to be used when accessing a file, pipe, or printer
    """

    FILE_READ_DATA = 0x00000001
    FILE_WRITE_DATA = 0x00000002
    FILE_APPEND_DATA = 0x00000004
    FILE_READ_EA = 0x00000008
    FILE_WRITE_EA = 0x00000010
    FILE_DELETE_CHILD = 0x00000040
    FILE_EXECUTE = 0x00000020
    FILE_READ_ATTRIBUTES = 0x00000080
    FILE_WRITE_ATTRIBUTES = 0x00000100
    DELETE = 0x00010000
    READ_CONTROL = 0x00020000
    WRITE_DAC = 0x00040000
    WRITE_OWNER = 0x00080000
    SYNCHRONIZE = 0x00100000
    ACCESS_SYSTEM_SECURITY = 0x01000000
    MAXIMUM_ALLOWED = 0x02000000
    GENERIC_ALL = 0x10000000
    GENERIC_EXECUTE = 0x20000000
    GENERIC_WRITE = 0x40000000
    GENERIC_READ = 0x80000000


class DirectoryAccessMask:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.13.1.2 Directory_Access_Mask
    Access Mask flag values to be used when accessing a directory
    """

    FILE_LIST_DIRECTORY = 0x00000001
    FILE_ADD_FILE = 0x00000002
    FILE_ADD_SUBDIRECTORY = 0x00000004
    FILE_READ_EA = 0x00000008
    FILE_WRITE_EA = 0x00000010
    FILE_TRAVERSE = 0x00000020
    FILE_DELETE_CHILD = 0x00000040
    FILE_READ_ATTRIBUTES = 0x00000080
    FILE_WRITE_ATTRIBUTES = 0x00000100
    DELETE = 0x00010000
    READ_CONTROL = 0x00020000
    WRITE_DAC = 0x00040000
    WRITE_OWNER = 0x00080000
    SYNCHRONIZE = 0x00100000
    ACCESS_SYSTEM_SECURITY = 0x01000000
    MAXIMUM_ALLOWED = 0x02000000
    GENERIC_ALL = 0x10000000
    GENERIC_EXECUTE = 0x20000000
    GENERIC_WRITE = 0x40000000
    GENERIC_READ = 0x80000000


class FileFlags:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.14 SMB2 CREATE Response Flags
    Flag that details info about the file that was opened.
    """

    SMB2_CREATE_FLAG_REPARSEPOINT = 0x1


class CreateAction:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.14 SMB2 CREATE Response Flags
    The action taken in establishing the open.
    """

    FILE_SUPERSEDED = 0x0
    FILE_OPENED = 0x1
    FILE_CREATED = 0x2
    FILE_OVERWRITTEN = 0x3


class CloseFlags:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.15 SMB2 CLOSE Request Flags
    Flags to indicate how to process the operation
    """

    SMB2_CLOSE_FLAG_POSTQUERY_ATTRIB = 0x01


class ReadFlags:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.19 SMB2 READ Request Flags
    Read flags for SMB 3.0.2 and newer dialects
    """

    SMB2_READFLAG_READ_UNBUFFERED = 0x01


class ReadWriteChannel:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.19/21 SMB2 READ/Write Request Channel
    Channel information for an SMB2 READ Request message
    """

    SMB2_CHANNEL_NONE = 0x0
    SMB2_CHANNEL_RDMA_V1 = 0x1
    SMB2_CHANNEL_RDMA_V1_INVALIDATE = 0x2


class WriteFlags:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.21 SMB2 WRITE Request Flags
    Flags to indicate how to process the operation
    """

    SMB2_WRITEFLAG_WRITE_THROUGH = 0x00000001
    SMB2_WRITEFLAG_WRITE_UNBUFFERED = 0x00000002


class QueryDirectoryFlags:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.33 SMB2 QUERY_DIRECTORY Request Flags
    Indicates how the query directory operation MUST be processed.
    """

    SMB2_RESTART_SCANS = 0x01
    SMB2_RETURN_SINGLE_ENTRY = 0x02
    SMB2_INDEX_SPECIFIED = 0x04
    SMB2_REOPEN = 0x10


class QueryInfoFlags:
    """
    [MS-SMB2] 2.2.37 SMB2 QUERY_INFO Request - Flags
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/d623b2f7-a5cd-4639-8cc9-71fa7d9f9ba9

    Flags to set on a QUERY_INFO request.
    """

    SL_RESTART_SCAN = 0x00000001
    SL_RETURN_SINGLE_ENTRY = 0x00000002
    SL_INDEX_SPECIFIED = 0x00000004


class InfoAdditionalInformation:
    """
    [MS-SMB2] 2.2.39 SMB2 SET_INFO Request - AdditionalInformation
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/ee9614c4-be54-4a3c-98f1-769a7032a0e4

    If security information is being set, this value must contains one or more of the following flags.
    """

    OWNER_SECURTIY_INFORMATION = 0x00000001
    GROUP_SECURITY_INFORMATION = 0x00000002
    DACL_SECURITY_INFORMATION = 0x00000004
    SACL_SECURITY_INFORMATION = 0x00000008
    LABEL_SECURITY_INFORMATION = 0x00000010
    ATTRIBUTE_SECURITY_INFORMATION = 0x00000020
    SCOPE_SECURITY_INFORMATION = 0x00000040
    BACKUP_SECURITY_INFORMATION = 0x00010000


class SMB2CreateRequest(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.13 SMB2 CREATE Request
    The SMB2 Create Request packet is sent by a client to request either
    creation of or access to a file.
    """

    COMMAND = Commands.SMB2_CREATE

    def __init__(self):
        # pep 80 char issues force me to define this here
        create_con_req = SMB2CreateContextRequest
        self.fields = OrderedDict(
            [
                (
                    "structure_size",
                    IntField(
                        size=2,
                        default=57,
                    ),
                ),
                ("security_flags", IntField(size=1)),
                ("requested_oplock_level", EnumField(size=1, enum_type=RequestedOplockLevel)),
                ("impersonation_level", EnumField(size=4, enum_type=ImpersonationLevel)),
                ("smb_create_flags", IntField(size=8)),
                ("reserved", IntField(size=8)),
                ("desired_access", IntField(size=4)),
                ("file_attributes", IntField(size=4)),
                ("share_access", FlagField(size=4, flag_type=ShareAccess)),
                ("create_disposition", EnumField(size=4, enum_type=CreateDisposition)),
                ("create_options", FlagField(size=4, flag_type=CreateOptions)),
                ("name_offset", IntField(size=2, default=120)),  # (header size 64) + (structure size 56)
                ("name_length", IntField(size=2, default=lambda s: self._name_length(s))),
                ("create_contexts_offset", IntField(size=4, default=lambda s: self._create_contexts_offset(s))),
                ("create_contexts_length", IntField(size=4, default=lambda s: len(s["buffer_contexts"]))),
                # Technically these are all under buffer but we split it to make
                # things easier
                (
                    "buffer_path",
                    BytesField(
                        size=lambda s: self._buffer_path_size(s),
                    ),
                ),
                (
                    "padding",
                    BytesField(size=lambda s: self._padding_size(s), default=lambda s: b"\x00" * self._padding_size(s)),
                ),
                (
                    "buffer_contexts",
                    ListField(
                        size=lambda s: s["create_contexts_length"].get_value(),
                        list_type=StructureField(structure_type=create_con_req),
                        unpack_func=lambda s, d: self._buffer_context_list(s, d),
                    ),
                ),
            ]
        )
        super().__init__()

    def _name_length(self, structure):
        buffer_path = structure["buffer_path"].get_value()
        return len(buffer_path) if buffer_path != b"\x00\x00" else 0

    def _create_contexts_offset(self, structure):
        if len(structure["buffer_contexts"]) == 0:
            return 0
        else:
            return structure["name_offset"].get_value() + len(structure["padding"]) + len(structure["buffer_path"])

    def _buffer_path_size(self, structure):
        name_length = structure["name_length"].get_value()
        return name_length if name_length != 0 else 2

    def _padding_size(self, structure):
        # no padding is needed if there are no contexts
        if structure["create_contexts_length"].get_value() == 0:
            return 0

        mod = structure["name_length"].get_value() % 8
        return 0 if mod == 0 else 8 - mod

    def _buffer_context_list(self, structure, data):
        context_list = []
        last_context = data == b""
        while not last_context:
            create_context = SMB2CreateContextRequest()
            data = create_context.unpack(data)
            context_list.append(create_context)
            last_context = create_context["next"].get_value() == 0

        return context_list


class SMB2CreateResponse(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.14 SMB2 CREATE Response
    The SMB2 Create Response packet is sent by the server to an SMB2 CREATE
    Request.
    """

    COMMAND = Commands.SMB2_CREATE

    def __init__(self):
        create_con_req = SMB2CreateContextRequest
        self.fields = OrderedDict(
            [
                ("structure_size", IntField(size=2, default=89)),
                ("oplock_level", EnumField(size=1, enum_type=RequestedOplockLevel)),
                ("flag", FlagField(size=1, flag_type=FileFlags)),
                ("create_action", EnumField(size=4, enum_type=CreateAction)),
                ("creation_time", DateTimeField(size=8)),
                ("last_access_time", DateTimeField(size=8)),
                ("last_write_time", DateTimeField(size=8)),
                ("change_time", DateTimeField(size=8)),
                ("allocation_size", IntField(size=8)),
                ("end_of_file", IntField(size=8)),
                (
                    "file_attributes",
                    FlagField(
                        size=4,
                        flag_type=FileAttributes,
                        flag_strict=False,
                    ),
                ),
                ("reserved2", IntField(size=4)),
                ("file_id", BytesField(size=16)),
                ("create_contexts_offset", IntField(size=4, default=lambda s: self._create_contexts_offset(s))),
                ("create_contexts_length", IntField(size=4, default=lambda s: len(s["buffer"]))),
                (
                    "buffer",
                    ListField(
                        size=lambda s: s["create_contexts_length"].get_value(),
                        list_type=StructureField(structure_type=create_con_req),
                        unpack_func=lambda s, d: self._buffer_context_list(s, d),
                    ),
                ),
            ]
        )
        super().__init__()

    def _create_contexts_offset(self, structure):
        if len(structure["buffer"]) == 0:
            return 0
        else:
            return 152

    def _buffer_context_list(self, structure, data):
        context_list = []
        last_context = data == b""
        while not last_context:
            create_context = SMB2CreateContextRequest()
            data = create_context.unpack(data)
            context_list.append(create_context)
            # Manually make sure the final padding is present
            create_context["padding2"] = b"\x00" * create_context._padding2_size(create_context)
            last_context = create_context["next"].get_value() == 0

        return context_list


class SMB2CloseRequest(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.15 SMB2 CLOSE Request
    Used by the client to close an instance of a file
    """

    COMMAND = Commands.SMB2_CLOSE

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("structure_size", IntField(size=2, default=24)),
                ("flags", FlagField(size=2, flag_type=CloseFlags)),
                ("reserved", IntField(size=4)),
                ("file_id", BytesField(size=16)),
            ]
        )
        super().__init__()


class SMB2CloseResponse(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.16 SMB2 CLOSE Response
    The response of a SMB2 CLOSE Request
    """

    COMMAND = Commands.SMB2_CLOSE

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("structure_size", IntField(size=2, default=60)),
                ("flags", FlagField(size=2, flag_type=CloseFlags)),
                ("reserved", IntField(size=4)),
                ("creation_time", DateTimeField()),
                ("last_access_time", DateTimeField()),
                ("last_write_time", DateTimeField()),
                ("change_time", DateTimeField()),
                ("allocation_size", IntField(size=8)),
                ("end_of_file", IntField(size=8)),
                (
                    "file_attributes",
                    FlagField(
                        size=4,
                        flag_type=FileAttributes,
                        flag_strict=False,
                    ),
                ),
            ]
        )
        super().__init__()


class SMB2FlushRequest(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.17 SMB2 FLUSH Request
    Flush all cached file information for a specified open of a file to the
    persistent store that backs the file.
    """

    COMMAND = Commands.SMB2_FLUSH

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("structure_size", IntField(size=2, default=24)),
                ("reserved1", IntField(size=2)),
                ("reserved2", IntField(size=4)),
                ("file_id", BytesField(size=16)),
            ]
        )
        super().__init__()


class SMB2FlushResponse(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.18 SMB2 FLUSH Response
    SMB2 FLUSH Response packet sent by the server.
    """

    COMMAND = Commands.SMB2_FLUSH

    def __init__(self):
        self.fields = OrderedDict([("structure_size", IntField(size=2, default=4)), ("reserved", IntField(size=2))])
        super().__init__()


class SMB2ReadRequest(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.19 SMB2 READ Request
    The request is used to run a read operation on the file specified.
    """

    COMMAND = Commands.SMB2_READ

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("structure_size", IntField(size=2, default=49)),
                ("padding", IntField(size=1)),
                ("flags", FlagField(size=1, flag_type=ReadFlags)),
                ("length", IntField(size=4)),
                ("offset", IntField(size=8)),
                ("file_id", BytesField(size=16)),
                ("minimum_count", IntField(size=4)),
                ("channel", FlagField(size=4, flag_type=ReadWriteChannel)),
                ("remaining_bytes", IntField(size=4)),
                ("read_channel_info_offset", IntField(size=2, default=lambda s: self._get_read_channel_info_offset(s))),
                ("read_channel_info_length", IntField(size=2, default=lambda s: self._get_read_channel_info_length(s))),
                ("buffer", BytesField(size=lambda s: self._get_buffer_length(s), default=b"\x00")),
            ]
        )
        super().__init__()

    def _get_read_channel_info_offset(self, structure):
        if structure["channel"].get_value() == 0:
            return 0
        else:
            return 64 + structure["structure_size"].get_value() - 1

    def _get_read_channel_info_length(self, structure):
        if structure["channel"].get_value() == 0:
            return 0
        else:
            return len(structure["buffer"].get_value())

    def _get_buffer_length(self, structure):
        # buffer should contain 1 byte of \x00 and not be empty
        if structure["channel"].get_value() == 0:
            return 1
        else:
            return structure["read_channel_info_length"].get_value()


class SMB2ReadResponse(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.20 SMB2 READ Response
    Response to an SMB2 READ Request.
    """

    COMMAND = Commands.SMB2_READ

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("structure_size", IntField(size=2, default=17)),
                ("data_offset", IntField(size=1)),
                ("reserved", IntField(size=1)),
                ("data_length", IntField(size=4, default=lambda s: len(s["buffer"]))),
                ("data_remaining", IntField(size=4)),
                ("reserved2", IntField(size=4)),
                ("buffer", BytesField(size=lambda s: s["data_length"].get_value())),
            ]
        )
        super().__init__()


class SMB2WriteRequest(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.21 SMB2 WRITE Request
    A write packet to sent to an open file or named pipe on the server
    """

    COMMAND = Commands.SMB2_WRITE

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("structure_size", IntField(size=2, default=49)),
                (
                    "data_offset",
                    IntField(size=2, default=0x70),  # offset to the buffer field  # seems to be hardcoded to this value
                ),
                ("length", IntField(size=4, default=lambda s: len(s["buffer"]))),
                ("offset", IntField(size=8)),  # the offset in the file of the data
                ("file_id", BytesField(size=16)),
                ("channel", FlagField(size=4, flag_type=ReadWriteChannel)),
                ("remaining_bytes", IntField(size=4)),
                (
                    "write_channel_info_offset",
                    IntField(size=2, default=lambda s: self._get_write_channel_info_offset(s)),
                ),
                ("write_channel_info_length", IntField(size=2, default=lambda s: len(s["buffer_channel_info"]))),
                ("flags", FlagField(size=4, flag_type=WriteFlags)),
                ("buffer", BytesField(size=lambda s: s["length"].get_value())),
                ("buffer_channel_info", BytesField(size=lambda s: s["write_channel_info_length"].get_value())),
            ]
        )
        super().__init__()

    def _get_write_channel_info_offset(self, structure):
        if len(structure["buffer_channel_info"]) == 0:
            return 0
        else:
            header_size = 64
            packet_size = structure["structure_size"].get_value() - 1
            buffer_size = len(structure["buffer"])
            return header_size + packet_size + buffer_size


class SMB2WriteResponse(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.22 SMB2 WRITE Response
    The response to the SMB2 WRITE Request sent by the server
    """

    COMMAND = Commands.SMB2_WRITE

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("structure_size", IntField(size=2, default=17)),
                ("reserved", IntField(size=2)),
                ("count", IntField(size=4)),
                ("remaining", IntField(size=4)),
                ("write_channel_info_offset", IntField(size=2)),
                ("write_channel_info_length", IntField(size=2)),
            ]
        )
        super().__init__()


class SMB2QueryDirectoryRequest(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.33 QUERY_DIRECTORY Request
    Used by the client to obtain a directory enumeration on a directory open.
    """

    COMMAND = Commands.SMB2_QUERY_DIRECTORY

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("structure_size", IntField(size=2, default=33)),
                ("file_information_class", EnumField(size=1, enum_type=FileInformationClass)),
                ("flags", FlagField(size=1, flag_type=QueryDirectoryFlags)),
                ("file_index", IntField(size=4)),
                ("file_id", BytesField(size=16)),
                ("file_name_offset", IntField(size=2, default=lambda s: 0 if len(s["buffer"]) == 0 else 96)),
                ("file_name_length", IntField(size=2, default=lambda s: len(s["buffer"]))),
                ("output_buffer_length", IntField(size=4)),
                # UTF-16-LE encoded search pattern
                ("buffer", BytesField(size=lambda s: s["file_name_length"].get_value())),
            ]
        )
        super().__init__()

    @staticmethod
    def unpack_response(file_information_class, buffer):
        """
        Pass in the buffer value from the response object to unpack it and
        return a list of query response structures for the request.

        :param file_information_class: The info class that represents the
            buffer.
        :param buffer: The raw bytes value of the SMB2QueryDirectoryResponse
            buffer field.
        :return: List of query_info.* structures based on the
            FileInformationClass used in the initial query request.
        """
        resp_structure = {
            FileInformationClass.FILE_DIRECTORY_INFORMATION: FileDirectoryInformation,
            FileInformationClass.FILE_NAMES_INFORMATION: FileNamesInformation,
            FileInformationClass.FILE_BOTH_DIRECTORY_INFORMATION: FileBothDirectoryInformation,
            FileInformationClass.FILE_ID_BOTH_DIRECTORY_INFORMATION: FileIdBothDirectoryInformation,
            FileInformationClass.FILE_FULL_DIRECTORY_INFORMATION: FileFullDirectoryInformation,
            FileInformationClass.FILE_ID_FULL_DIRECTORY_INFORMATION: FileIdFullDirectoryInformation,
        }[file_information_class]
        query_results = []

        current_offset = 0
        is_next = True
        while is_next:
            result = resp_structure()
            result.unpack(buffer[current_offset:])
            query_results.append(result)
            current_offset += result["next_entry_offset"].get_value()
            is_next = result["next_entry_offset"].get_value() != 0

        return query_results


class SMB2QueryDirectoryResponse(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.34 SMB2 QUERY_DIRECTORY Response
    Response to an SMB2 QUERY_DIRECTORY Request.
    """

    COMMAND = Commands.SMB2_QUERY_DIRECTORY

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("structure_size", IntField(size=2, default=9)),
                ("output_buffer_offset", IntField(size=2, default=72)),
                ("output_buffer_length", IntField(size=4, default=lambda s: len(s["buffer"]))),
                # this structure varies based on the requested information class
                ("buffer", BytesField(size=lambda s: s["output_buffer_length"].get_value())),
            ]
        )
        super().__init__()


class SMB2QueryInfoRequest(Structure):
    """
    [MS-SMB2] 2.2.37 SMB2 QUERY_INFO Request
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/d623b2f7-a5cd-4639-8cc9-71fa7d9f9ba9

    Sent by a client to request information on a file, named pipe, or underlying volume.
    """

    COMMAND = Commands.SMB2_QUERY_INFO

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "structure_size",
                    IntField(
                        size=2,
                        default=41,
                    ),
                ),
                (
                    "info_type",
                    EnumField(
                        size=1,
                        enum_type=InfoType,
                    ),
                ),
                (
                    "file_info_class",
                    EnumField(size=1, enum_type=FileInformationClass, default=FileInformationClass.FILE_NONE),
                ),
                (
                    "output_buffer_length",
                    IntField(
                        size=4,
                        default=lambda s: 0,
                    ),
                ),
                (
                    "input_buffer_offset",
                    IntField(
                        size=2,
                        default=lambda s: 0 if s["input_buffer_length"].get_value() == 0 else 104,
                    ),
                ),
                ("reserved", IntField(size=2)),
                (
                    "input_buffer_length",
                    IntField(
                        size=4,
                        default=lambda s: len(s["buffer"]),
                    ),
                ),
                ("additional_information", IntField(size=4)),
                (
                    "flags",
                    FlagField(
                        size=4,
                        flag_type=QueryInfoFlags,
                    ),
                ),
                ("file_id", BytesField(size=16)),
                (
                    "buffer",
                    BytesField(
                        size=lambda s: s["input_buffer_length"].get_value(),
                    ),
                ),
            ]
        )
        super().__init__()


class SMB2QueryInfoResponse(Structure):
    """
    [MS-SMB2] 2.2.38 SMB2 QUERY_INFO Response
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/3b1b3598-a898-44ca-bfac-2dcae065247f

    Sent by the server in response to an SMB2QueryInfoRequest.
    """

    COMMAND = Commands.SMB2_QUERY_INFO

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "structure_size",
                    IntField(
                        size=2,
                        default=9,
                    ),
                ),
                (
                    "output_buffer_offset",
                    IntField(
                        size=2,
                        default=72,
                    ),
                ),
                (
                    "output_buffer_length",
                    IntField(
                        size=4,
                        default=lambda s: len(s["buffer"]),
                    ),
                ),
                (
                    "buffer",
                    BytesField(
                        size=lambda s: s["output_buffer_length"].get_value(),
                    ),
                ),
            ]
        )
        super().__init__()

    def parse_buffer(self, file_info_type):
        buffer = self["buffer"].get_value()

        def unpack_list(buffer, byte_boundary):
            info_list = []
            while buffer:
                entry = file_info_type()
                buffer = entry.unpack(buffer)

                padded_size = len(entry) % byte_boundary
                buffer_offset = (byte_boundary - padded_size) if padded_size else 0
                buffer = buffer[buffer_offset:]

                info_list.append(entry)

            return info_list

        file_obj = file_info_type()
        if isinstance(file_obj, FileFullEaInformation):
            return unpack_list(buffer, 4)
        elif isinstance(file_obj, FileStreamInformation):
            return unpack_list(buffer, 8)
        else:
            file_obj.unpack(buffer)
            return file_obj


class SMB2SetInfoRequest(Structure):
    """
    [MS-SMB2] 2.2.39 SMB2 SET_INFO Request
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/ee9614c4-be54-4a3c-98f1-769a7032a0e4

    Sent by a client to set information on a file or underlying object store.
    """

    COMMAND = Commands.SMB2_SET_INFO

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "structure_size",
                    IntField(
                        size=2,
                        default=33,
                    ),
                ),
                (
                    "info_type",
                    EnumField(
                        size=1,
                        enum_type=InfoType,
                    ),
                ),
                (
                    "file_info_class",
                    EnumField(size=1, enum_type=FileInformationClass, default=FileInformationClass.FILE_NONE),
                ),
                (
                    "buffer_length",
                    IntField(
                        size=4,
                        default=lambda s: len(s["buffer"]),
                    ),
                ),
                ("buffer_offset", IntField(size=2, default=96)),
                ("reserved", IntField(size=2)),
                (
                    "additional_information",
                    FlagField(
                        size=4,
                        flag_type=InfoAdditionalInformation,
                    ),
                ),
                ("file_id", BytesField(size=16)),
                ("buffer", BytesField(size=lambda s: s["buffer_length"].get_value())),
            ]
        )
        super().__init__()


class SMB2SetInfoResponse(Structure):
    """
    [MS-SMB2] 2.2.40 SMB2 SET_INFO Response
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/c4318eb4-bdab-49b7-9352-abd7005c7f19

    Sent by the server in response to an SMB2SetInfoRequest to notify the client that its request has been successfully
    processed.
    """

    COMMAND = Commands.SMB2_SET_INFO

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("structure_size", IntField(size=2, default=2)),
            ]
        )
        super().__init__()


class Open:
    def __init__(self, tree, name):
        """
        [MS-SMB2] v53.0 2017-09-15

        3.2.1.6 Per Application Open of a File
        Attributes per each open of a file. A file can be a File, Pipe,
        Directory, or Printer

        :param tree: The Tree (share) the file is located in.
        :param name: The name of the file, excluding the share path.
        """
        # properties available based on the file itself
        self._connected = False
        self.create_action = None
        self.creation_time = None
        self.last_access_time = None
        self.last_write_time = None
        self.change_time = None
        self.allocation_size = None
        self.end_of_file = None
        self.file_attributes = None

        # properties used privately
        # set to { 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF } to allow message
        # compounding with the open as the first message, once opened this
        # will be overwritten by the open response
        self.file_id = b"\xff" * 16
        self.tree_connect = tree
        self.connection = tree.session.connection
        self.oplock_level = None
        self.durable = None
        self.file_name = name
        self.resilient_handle = None
        self.last_disconnect_time = None
        self.resilient_timeout = None

        # an array of entries used to maintain information about outstanding
        # lock and unlock operations performed on resilient Opens. Contains
        #     sequence_number - 4-bit integer modulo 16
        #     free - boolean value where False is no outstanding requests
        self.operation_buckets = []

        # SMB 3.x+
        self.durable_timeout = None

        # Table of outstanding requests, lookup by Request.cancel_id,
        # message_id
        self.outstanding_requests = {}

        self.create_guid = None
        self.is_persistent = None
        self.desired_access = None
        self.share_mode = None
        self.create_options = None
        self.file_attributes = None
        self.create_disposition = None

    @property
    def connected(self):
        return self._connected

    def create(
        self,
        impersonation_level,
        desired_access,
        file_attributes,
        share_access,
        create_disposition,
        create_options,
        create_contexts=None,
        oplock_level=RequestedOplockLevel.SMB2_OPLOCK_LEVEL_NONE,
        send=True,
    ):
        """
        This will open the file based on the input parameters supplied. Any
        file open should also be called with Open.close() when it is finished.

        More details on how each option affects the open process can be found
        here https://msdn.microsoft.com/en-us/library/cc246502.aspx.

        Supports out of band send function, call this function with send=False
        to return a tuple of (SMB2CreateRequest, receive_func) instead of
        sending the the request and waiting for the response. The receive_func
        can be used to get the response from the server by passing in the
        Request that was used to sent it out of band.

        :param impersonation_level: (ImpersonationLevel) The type of
            impersonation level that is issuing the create request.
        :param desired_access: The level of access that is required of the
            open. FilePipePrinterAccessMask or DirectoryAccessMask should be
            used depending on the type of file being opened.
        :param file_attributes: (FileAttributes) attributes to set on the file
            being opened, this usually is for opens that creates a file.
        :param share_access: (ShareAccess) Specifies the sharing mode for the
            open.
        :param create_disposition: (CreateDisposition) Defines the action the
            server MUST take if the file already exists.
        :param create_options: (CreateOptions) Specifies the options to be
            applied when creating or opening the file.
        :param create_contexts: (List<SMB2CreateContextRequest>) List of
            context request values to be applied to the create.

        Create Contexts are used to encode additional flags and attributes when
        opening files. More details on create context request values can be
        found here https://msdn.microsoft.com/en-us/library/cc246504.aspx.

        :param oplock_level: The requested oplock level of the request.
        :param send: Whether to send the request in the same call or return the
            message to the caller and the unpack function

        :return: List of context response values or None if there are no
            context response values. If the context response value is not known
            to smbprotocol then the list value would be raw bytes otherwise
            it is a Structure defined in create_contexts.py
        """
        create = SMB2CreateRequest()
        create["requested_oplock_level"] = oplock_level
        create["impersonation_level"] = impersonation_level
        create["desired_access"] = desired_access
        create["file_attributes"] = file_attributes
        create["share_access"] = share_access
        create["create_disposition"] = create_disposition
        create["create_options"] = create_options
        if self.file_name == "":
            create["buffer_path"] = b"\x00\x00"
        else:
            create["buffer_path"] = self.file_name.encode("utf-16-le")
        if create_contexts:
            create["buffer_contexts"] = SMB2CreateContextRequest.pack_multiple(create_contexts)

        if self.connection.dialect >= Dialects.SMB_3_0_0:
            self.desired_access = desired_access
            self.share_mode = share_access
            self.create_options = create_options
            self.file_attributes = file_attributes
            self.create_disposition = create_disposition

        if not send:
            return create, self._create_response

        log.info(
            "Session: %s, Tree Connect: %s - sending SMB2 Create Request for file %s",
            self.tree_connect.session.username,
            self.tree_connect.share_name,
            self.file_name,
        )

        log.debug(create)
        request = self.connection.send(create, self.tree_connect.session.session_id, self.tree_connect.tree_connect_id)
        return self._create_response(request)

    def _create_response(self, request):
        log.info(
            "Session: %s, Tree Connect: %s - receiving SMB2 Create Response",
            self.tree_connect.session.username,
            self.tree_connect.share_name,
        )
        response = self.connection.receive(request)
        create_response = SMB2CreateResponse()
        create_response.unpack(response["data"].get_value())

        # Manually set the length so the debug log won't fail due to some servers returning a padded value which is not
        # reflected in the padding2 of the context response.
        create_response["create_contexts_length"] = len(create_response["buffer"])

        self._connected = True
        log.debug(create_response)

        self.file_id = create_response["file_id"].get_value()
        self.tree_connect.session.open_table[self.file_id] = self
        self.oplock_level = create_response["oplock_level"].get_value()
        self.durable = False
        self.resilient_handle = False
        self.last_disconnect_time = 0

        self.create_action = create_response["create_action"].get_value()
        self.creation_time = create_response["creation_time"].get_value()
        self.last_access_time = create_response["last_access_time"].get_value()
        self.last_write_time = create_response["last_write_time"].get_value()
        self.change_time = create_response["change_time"].get_value()
        self.allocation_size = create_response["allocation_size"].get_value()
        self.end_of_file = create_response["end_of_file"].get_value()
        self.file_attributes = create_response["file_attributes"].get_value()

        create_contexts_response = None
        if create_response["create_contexts_length"].get_value() > 0:
            create_contexts_response = []
            for context in create_response["buffer"].get_value():
                create_contexts_response.append(context.get_context_data())

        return create_contexts_response

    def read(self, offset, length, min_length=0, unbuffered=False, wait=True, send=True):
        """
        Reads from an opened file or pipe

        Supports out of band send function, call this function with send=False
        to return a tuple of (SMB2ReadRequest, receive_func) instead of
        sending the the request and waiting for the response. The receive_func
        can be used to get the response from the server by passing in the
        Request that was used to sent it out of band.

        :param offset: The offset to start the read of the file.
        :param length: The number of bytes to read from the offset.
        :param min_length: The minimum number of bytes to be read for a
            successful operation.
        :param unbuffered: Whether to the server should cache the read data at
            intermediate layers, only value for SMB 3.0.2 or newer
        :param wait: If send=True, whether to wait for a response if
            STATUS_PENDING was received from the server or fail.
        :param send: Whether to send the request in the same call or return the
            message to the caller and the unpack function
        :return: A byte string of the bytes read
        """
        if length > self.connection.max_read_size:
            raise SMBException(
                f"The requested read length {length} is greater than "
                f"the maximum negotiated read size {self.connection.max_read_size}"
            )

        read = SMB2ReadRequest()
        read["length"] = length
        read["offset"] = offset
        read["minimum_count"] = min_length
        read["file_id"] = self.file_id
        read["padding"] = b"\x50"

        if unbuffered:
            if self.connection.dialect < Dialects.SMB_3_0_2:
                raise SMBUnsupportedFeature(
                    self.connection.dialect, Dialects.SMB_3_0_2, "SMB2_READFLAG_READ_UNBUFFERED", True
                )
            read["flags"].set_flag(ReadFlags.SMB2_READFLAG_READ_UNBUFFERED)

        if not send:
            return read, self._read_response

        log.info(
            "Session: %s, Tree Connect ID: %s - sending SMB2 Read Request for file %s",
            self.tree_connect.session.username,
            self.tree_connect.share_name,
            self.file_name,
        )
        log.debug(read)
        request = self.connection.send(read, self.tree_connect.session.session_id, self.tree_connect.tree_connect_id)
        return self._read_response(request, wait)

    def _read_response(self, request, wait=True):
        log.info(
            "Session: %s, Tree Connect ID: %s - receiving SMB2 Read Response",
            self.tree_connect.session.username,
            self.tree_connect.share_name,
        )
        response = self.connection.receive(request, wait=wait)
        read_response = SMB2ReadResponse()
        read_response.unpack(response["data"].get_value())
        log.debug(read_response)

        return read_response["buffer"].get_value()

    def write(self, data, offset=0, write_through=False, unbuffered=False, wait=True, send=True):
        """
        Writes data to an opened file.

        Supports out of band send function, call this function with send=False
        to return a tuple of (SMBWriteRequest, receive_func) instead of
        sending the the request and waiting for the response. The receive_func
        can be used to get the response from the server by passing in the
        Request that was used to sent it out of band.

        :param data: The bytes data to write.
        :param offset: The offset in the file to write the bytes at
        :param write_through: Whether written data is persisted to the
            underlying storage, not valid for SMB 2.0.2.
        :param unbuffered: Whether to the server should cache the write data at
            intermediate layers, only value for SMB 3.0.2 or newer
        :param wait: If send=True, whether to wait for a response if
            STATUS_PENDING was received from the server or fail.
        :param send: Whether to send the request in the same call or return the
            message to the caller and the unpack function
        :return: The number of bytes written
        """
        data_len = len(data)
        if data_len > self.connection.max_write_size:
            raise SMBException(
                f"The requested write length {data_len} is greater than "
                f"the maximum negotiated write size {self.connection.max_write_size}"
            )

        write = SMB2WriteRequest()
        write["length"] = len(data)
        write["offset"] = offset
        write["file_id"] = self.file_id
        write["buffer"] = data

        if write_through:
            if self.connection.dialect < Dialects.SMB_2_1_0:
                raise SMBUnsupportedFeature(
                    self.connection.dialect, Dialects.SMB_2_1_0, "SMB2_WRITEFLAG_WRITE_THROUGH", True
                )
            write["flags"].set_flag(WriteFlags.SMB2_WRITEFLAG_WRITE_THROUGH)

        if unbuffered:
            if self.connection.dialect < Dialects.SMB_3_0_2:
                raise SMBUnsupportedFeature(
                    self.connection.dialect, Dialects.SMB_3_0_2, "SMB2_WRITEFLAG_WRITE_UNBUFFERED", True
                )
            write["flags"].set_flag(WriteFlags.SMB2_WRITEFLAG_WRITE_UNBUFFERED)

        if not send:
            return write, self._write_response

        log.info(
            "Session: %s, Tree Connect: %s - sending SMB2 Write Request for file %s",
            self.tree_connect.session.username,
            self.tree_connect.share_name,
            self.file_name,
        )
        log.debug(write)
        request = self.connection.send(write, self.tree_connect.session.session_id, self.tree_connect.tree_connect_id)
        return self._write_response(request, wait)

    def _write_response(self, request, wait=True):
        log.info(
            "Session: %s, Tree Connect: %s - receiving SMB2 Write Response",
            self.tree_connect.session.username,
            self.tree_connect.share_name,
        )
        response = self.connection.receive(request, wait=wait)
        write_response = SMB2WriteResponse()
        write_response.unpack(response["data"].get_value())
        log.debug(write_response)

        return write_response["count"].get_value()

    def flush(self, send=True):
        """
        A command sent by the client to request that a server flush all cached
        file information for the opened file.

        Supports out of band send function, call this function with send=False
        to return a tuple of (SMB2FlushRequest, receive_func) instead of
        sending the the request and waiting for the response. The receive_func
        can be used to get the response from the server by passing in the
        Request that was used to sent it out of band.

        :param send: Whether to send the request in the same call or return the
            message to the caller and the unpack function
        :return: The SMB2FlushResponse received from the server
        """
        flush = SMB2FlushRequest()
        flush["file_id"] = self.file_id

        if not send:
            return flush, self._flush_response

        log.info(
            "Session: %s, Tree Connect: %s - sending SMB2 Flush Request for file %s",
            self.tree_connect.session.username,
            self.tree_connect.share_name,
            self.file_name,
        )
        log.debug(flush)
        request = self.connection.send(flush, self.tree_connect.session.session_id, self.tree_connect.tree_connect_id)
        return self._flush_response(request)

    def _flush_response(self, request):
        log.info(
            "Session: %s, Tree Connect: %s - receiving SMB2 Flush Response",
            self.tree_connect.session.username,
            self.tree_connect.share_name,
        )
        response = self.connection.receive(request)
        flush_response = SMB2FlushResponse()
        flush_response.unpack(response["data"].get_value())
        log.debug(flush_response)
        return flush_response

    def query_directory(
        self, pattern, file_information_class, flags=None, file_index=0, max_output=MAX_PAYLOAD_SIZE, send=True
    ):
        """
        Run a Query/Find on an opened directory based on the params passed in.

        Supports out of band send function, call this function with send=False
        to return a tuple of (SMB2QueryDirectoryRequest, receive_func) instead
        of sending the the request and waiting for the response. The
        receive_func can be used to get the response from the server by passing
        in the Request that was used to sent it out of band.

        :param pattern: The string pattern to use for the query, this pattern
            format is based on the SMB server but * is usually a wildcard
        :param file_information_class: FileInformationClass that defines the
            format of the result that is returned
        :param flags: QueryDirectoryFlags that control how the operation must
            be processed.
        :param file_index: If the flags SMB2_INDEX_SPECIFIED, this is the index
            the query should resume on, otherwise should be 0
        :param max_output: The maximum output size, defaulted to the max credit
            size but can be increased to reduced round trip operations.
        :param send: Whether to send the request in the same call or return the
            message to the caller and the unpack function
        :return: A list of structures defined in query_info.py, the list entry
            structure is based on the value of file_information_class in the
            request message
        """
        query = SMB2QueryDirectoryRequest()
        query["file_information_class"] = file_information_class
        query["flags"] = flags
        query["file_index"] = file_index
        query["file_id"] = self.file_id
        query["output_buffer_length"] = max_output
        query["buffer"] = pattern.encode("utf-16-le")

        if not send:
            return query, self._query_directory_response

        log.info(
            "Session: %s, Tree Connect: %s - sending SMB2 Query Directory Request for directory %s",
            self.tree_connect.session.username,
            self.tree_connect.share_name,
            self.file_name,
        )
        log.debug(query)
        request = self.connection.send(query, self.tree_connect.session.session_id, self.tree_connect.tree_connect_id)
        return self._query_directory_response(request)

    def _query_directory_response(self, request):
        log.info(
            "Session: %s, Tree Connect: %s - receiving SMB2 Query Response",
            self.tree_connect.session.username,
            self.tree_connect.share_name,
        )
        response = self.connection.receive(request)
        query_response = SMB2QueryDirectoryResponse()
        query_response.unpack(response["data"].get_value())
        log.debug(query_response)

        query_request = SMB2QueryDirectoryRequest()
        query_request.unpack(request.message["data"].get_value())
        file_cl = query_request["file_information_class"].get_value()
        data = query_response["buffer"].get_value()
        results = SMB2QueryDirectoryRequest.unpack_response(file_cl, data)
        return results

    def close(self, get_attributes=False, send=True):
        """
        Closes an opened file.

        Supports out of band send function, call this function with send=False
        to return a tuple of (SMB2CloseRequest, receive_func) instead of
        sending the the request and waiting for the response. The receive_func
        can be used to get the response from the server by passing in the
        Request that was used to sent it out of band.

        :param get_attributes: (Bool) whether to get the latest attributes on
            the close and set them on the Open object
        :param send: Whether to send the request in the same call or return the
            message to the caller and the unpack function
        :return: SMB2CloseResponse message received from the server
        """
        # it is already closed and this isn't for an out of band request
        if not self._connected and send:
            return

        close = SMB2CloseRequest()

        close["file_id"] = self.file_id
        if get_attributes:
            close["flags"] = CloseFlags.SMB2_CLOSE_FLAG_POSTQUERY_ATTRIB

        if not send:
            return close, self._close_response

        log.info(
            "Session: %s, Tree Connect: %s - sending SMB2 Close Request for file %s",
            self.tree_connect.session.username,
            self.tree_connect.share_name,
            self.file_name,
        )
        log.debug(close)
        request = self.connection.send(close, self.tree_connect.session.session_id, self.tree_connect.tree_connect_id)
        return self._close_response(request)

    def _close_response(self, request):
        log.info(
            "Session: %s, Tree Connect: %s - receiving SMB2 Close Response",
            self.tree_connect.session.username,
            self.tree_connect.share_name,
        )
        try:
            response = self.connection.receive(request)
        except FileClosed:
            self._connected = False
            self.tree_connect.session.open_table.pop(self.file_id, None)
            return

        c_resp = SMB2CloseResponse()
        c_resp.unpack(response["data"].get_value())
        log.debug(c_resp)
        self._connected = False
        del self.tree_connect.session.open_table[self.file_id]

        # update the attributes if requested
        close_request = SMB2CloseRequest()
        close_request.unpack(request.message["data"].get_value())
        if close_request["flags"].has_flag(CloseFlags.SMB2_CLOSE_FLAG_POSTQUERY_ATTRIB):
            self.creation_time = c_resp["creation_time"].get_value()
            self.last_access_time = c_resp["last_access_time"].get_value()
            self.last_write_time = c_resp["last_write_time"].get_value()
            self.change_time = c_resp["change_time"].get_value()
            self.allocation_size = c_resp["allocation_size"].get_value()
            self.end_of_file = c_resp["end_of_file"].get_value()
            self.file_attributes = c_resp["file_attributes"].get_value()
        return c_resp
